/*
 * uPD77660 clock framework support
 *
 * Copyright (C) 2011  Renesas Electronics Corporation
 * Copyright (C) 2010 Magnus Damm
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/io.h>
#include <linux/clk.h>
#include <linux/sh_clk.h>
#include <mach/common.h>
#include <asm/clkdev.h>

/* CPG registers */
#define FRQCRA		0xe6150000
#define FRQCRB		0xe6150004
#define FRQCRC		0xe61500e0
#define PLLC01CR	0xe6150028

#define VCLKCR1		0xe6150008
#define VCLKCR2		0xe615000c
#define FMSICKCR	0xe6150010
#define FMSOCKCR	0xe6150014
#define FSIACKCR	0xe6150018
#define SUBCKCR		0xe6150080
#define SPUCKCR		0xe6150084
#define VOUCKCR		0xe6150088
#define USBCKCR		0xe615008c
#define HDMICKCR	0xe6150094
#define STPRCKCR	0xe615009c

#define MSTPSR0		0xe6150030
#define MSTPSR1		0xe6150038
#define MSTPSR2		0xe6150040
#define MSTPSR3		0xe6150048
#define MSTPSR4		0xe615004c
#define SMSTPCR0	0xe6150130
#define SMSTPCR1	0xe6150134
#define SMSTPCR2	0xe6150138
#define SMSTPCR3	0xe615013c
#define SMSTPCR4	0xe6150140

/* Fixed 32 KHz root clock from EXTALR pin */
static struct clk extalr_clk = {
	.rate	= 32768,
};

/*
 * 25MHz default rate for the EXTAL1 root input clock.
 * If needed, reset this with clk_set_rate() from the platform code.
 */
static struct clk extal1_clk = {
	.rate	= 25000000,
};

/*
 * 48MHz default rate for the EXTAL2 root input clock.
 * If needed, reset this with clk_set_rate() from the platform code.
 */
static struct clk extal2_clk = {
	.rate	= 48000000,
};

/*
 * 27MHz default rate for the DV_CLKI root input clock.
 * If needed, reset this with clk_set_rate() from the platform code.
 */
static struct clk dv_clk = {
	.rate	= 27000000,
};

static unsigned long div_recalc(struct clk *clk)
{
	return clk->parent->rate / (int)(clk->priv);
}

static struct clk_ops div_clk_ops = {
	.recalc	= div_recalc,
};

/* Divide extal1 by two */
static struct clk extal1_div2_clk = {
	.ops	= &div_clk_ops,
	.priv	= (void *)2,
	.parent	= &extal1_clk,
};

/* Divide extal1 by 1024 */
static struct clk extal1_div1024_clk = {
	.ops	= &div_clk_ops,
	.priv	= (void *)1024,
	.parent	= &extal1_clk,
};

/* Divide extal2 by two */
static struct clk extal2_div2_clk = {
	.ops	= &div_clk_ops,
	.priv	= (void *)2,
	.parent	= &extal2_clk,
};

static struct clk_ops followparent_clk_ops = {
	.recalc	= followparent_recalc,
};

/* Main clock */
static struct clk system_clk = {
	.ops	= &followparent_clk_ops,
};

static struct clk system_div2_clk = {
	.ops	= &div_clk_ops,
	.priv	= (void *)2,
	.parent	= &system_clk,
};

/* r_clk */
static struct clk r_clk = {
	.ops	= &followparent_clk_ops,
};

/* PLLC0 and PLLC1 */
static unsigned long pllc01_recalc(struct clk *clk)
{
	unsigned long mult = 1;

	if (__raw_readl(PLLC01CR) & (1 << 14))
		mult = ((__raw_readl(clk->enable_reg) >> 24) & 0x7f) + 1;

	return clk->parent->rate * mult;
}

static struct clk_ops pllc01_clk_ops = {
	.recalc		= pllc01_recalc,
};

static struct clk pllc0_clk = {
	.ops		= &pllc01_clk_ops,
	.flags		= CLK_ENABLE_ON_INIT,
	.parent		= &system_clk,
	.enable_reg	= (void __iomem *)FRQCRC,
};

static struct clk pllc1_clk = {
	.ops		= &pllc01_clk_ops,
	.flags		= CLK_ENABLE_ON_INIT,
	.parent		= &system_div2_clk,
	.enable_reg	= (void __iomem *)FRQCRA,
};

/* Divide PLLC1 by two */
static struct clk pllc1_div2_clk = {
	.ops		= &div_clk_ops,
	.priv		= (void *)2,
	.parent		= &pllc1_clk,
};

/* USB24S clock */
static unsigned long usb24s_recalc(struct clk *clk)
{
	if (__raw_readl(USBCKCR) & (1 << 7))
		return extal2_clk.rate;
	else
		return clk_get_rate(&system_clk);
};

static struct clk_ops usb24s_clk_ops = {
	.recalc	= usb24s_recalc,
};

static struct clk usb24s_clk = {
	.ops	= &usb24s_clk_ops,
};

static struct clk usb24s_div2_clk = {
	.ops	= &div_clk_ops,
	.priv	= (void *)2,
	.parent	= &usb24s_clk,
};

/* External input clock (pin name: FMSICK/FMSOCK/FSIACK) */
struct clk fmsick_clk = {
};

struct clk fmsock_clk = {
};

struct clk fsiack_clk = {
};

/* CA9 clock */
static struct clk z_clk = {
	.ops	= &div_clk_ops,
	.parent = &pllc0_clk,
};

struct clk *main_clks[] = {
	&extalr_clk,
	&extal1_clk,
	&extal2_clk,
	&extal1_div2_clk,
	&extal1_div1024_clk,
	&extal2_div2_clk,
	&dv_clk,
	&system_clk,
	&system_div2_clk,
	&r_clk,
	&pllc0_clk,
	&pllc1_clk,
	&pllc1_div2_clk,
	&usb24s_clk,
	&usb24s_div2_clk,
	&fmsick_clk,
	&fmsock_clk,
	&fsiack_clk,
	&z_clk,
};

static void div4_kick(struct clk *clk)
{
	unsigned long value;

	/* set KICK bit in FRQCRB to update hardware setting */
	value = __raw_readl(FRQCRB);
	value |= (1 << 31);
	__raw_writel(value, FRQCRB);
}

static int divisors[] = { 2, 3, 4, 6, 8, 12, 16, 18,
			  24, 32, 36, 48, 0, 72, 96, 0 };

static struct clk_div_mult_table div4_div_mult_table = {
	.divisors = divisors,
	.nr_divisors = ARRAY_SIZE(divisors),
};

static struct clk_div4_table div4_table = {
	.div_mult_table = &div4_div_mult_table,
	.kick = div4_kick,
};

enum {	DIV4_I, DIV4_ZG, DIV4_B, DIV4_M1, DIV4_HP,
	DIV4_HPP, DIV4_USBP, DIV4_S, DIV4_ZB, DIV4_M3, DIV4_CP,
	DIV4_NR
};

#define DIV4(_reg, _bit, _mask, _flags) \
	SH_CLK_DIV4(&pllc1_clk, _reg, _bit, _mask, _flags)

struct clk div4_clks[DIV4_NR] = {
	[DIV4_I]	= DIV4(FRQCRA, 20, 0x6fff, CLK_ENABLE_ON_INIT),
	[DIV4_ZG]	= DIV4(FRQCRA, 16, 0x6fff, 0),
	[DIV4_B]	= DIV4(FRQCRA, 8, 0x6fff, CLK_ENABLE_ON_INIT),
	[DIV4_M1]	= DIV4(FRQCRA, 4, 0x6fff, CLK_ENABLE_ON_INIT),
	[DIV4_HP]	= DIV4(FRQCRB, 4, 0x6fff, 0),
	[DIV4_HPP]	= DIV4(FRQCRC, 20, 0x6fff, 0),
	[DIV4_USBP]	= DIV4(FRQCRC, 16, 0x6fff, 0),
	[DIV4_S]	= DIV4(FRQCRC, 12, 0x6fff, 0),
	[DIV4_ZB]	= DIV4(FRQCRC, 8, 0x6fff, 0),
	[DIV4_M3]	= DIV4(FRQCRC, 4, 0x6fff, 0),
	[DIV4_CP]	= DIV4(FRQCRC, 0, 0x6fff, 0),
};

enum {	DIV6_VCK1, DIV6_VCK2, DIV6_SPU, DIV6_STPR,
	DIV6_NR
};

static struct clk div6_clks[DIV6_NR] = {
	[DIV6_VCK1]	= SH_CLK_DIV6(&pllc1_div2_clk, VCLKCR1, 0),
	[DIV6_VCK2]	= SH_CLK_DIV6(&pllc1_div2_clk, VCLKCR2, 0),
	[DIV6_SPU]	= SH_CLK_DIV6(&pllc1_div2_clk, SPUCKCR, 0),
	[DIV6_STPR]	= SH_CLK_DIV6(&pllc1_div2_clk, STPRCKCR, 0),
};

enum {	DIV6_SUB, DIV6_SUB2, DIV6_VOU, DIV6_FMSI, DIV6_FMSO, DIV6_FSIA,
	DIV6_USB, DIV6_HDMI, DIV6_REPARENT_NR
};

/* Indices are important - they are the actual src selecting values */
static struct clk *subckcr_parent[] = {
	[0] = &pllc1_div2_clk,
	[1] = &usb24s_clk,
	[2] = NULL,	/* setting prohibited */
};

static struct clk *vouckcr_parent[] = {
	[0] = &pllc1_div2_clk,
	[1] = &system_clk,
	[2] = &dv_clk,
	[3] = NULL,	/* setting prohibited */
};

static struct clk *fmsickcr_parent[] = {
	[0] = &pllc1_div2_clk,
	[1] = &fmsick_clk, /* external input for FMSICK */
};

static struct clk *fmsockcr_parent[] = {
	[0] = &pllc1_div2_clk,
	[1] = &fmsock_clk, /* external input for FMSOCK */
};

static struct clk *fsiackcr_parent[] = {
	[0] = &pllc1_div2_clk,
	[1] = &fsiack_clk, /* external input for FSIACK */
};

static struct clk *usbckcr_parent[] = {
	[0] = &usb24s_div2_clk,
	[1] = &usb24s_clk,
};

static struct clk *hdmickcr_parent[] = {
	[0] = &pllc1_div2_clk,
	[1] = &extal1_clk,
	[2] = &dv_clk,
};

static struct clk div6_reparent_clks[DIV6_REPARENT_NR] = {
	[DIV6_SUB] = SH_CLK_DIV6_EXT(NULL, SUBCKCR, 9, CLK_DIV_SHARED,
			subckcr_parent, ARRAY_SIZE(subckcr_parent), 6, 1),
	[DIV6_SUB2] = SH_CLK_DIV6_EXT(NULL, SUBCKCR, 10, CLK_DIV_SHARED,
			subckcr_parent, ARRAY_SIZE(subckcr_parent), 6, 1),
	[DIV6_VOU] = SH_CLK_DIV6_EXT(NULL, VOUCKCR, 8, 0,
			vouckcr_parent, ARRAY_SIZE(vouckcr_parent), 6, 2),
	[DIV6_FMSI] = SH_CLK_DIV6_EXT(NULL, FMSICKCR, 8, 0,
			fmsickcr_parent, ARRAY_SIZE(fmsickcr_parent), 6, 1),
	[DIV6_FMSO] = SH_CLK_DIV6_EXT(NULL, FMSOCKCR, 8, 0,
			fmsockcr_parent, ARRAY_SIZE(fmsockcr_parent), 6, 1),
	[DIV6_FSIA] = SH_CLK_DIV6_EXT(NULL, FSIACKCR, 8, 0,
			fsiackcr_parent, ARRAY_SIZE(fsiackcr_parent), 6, 1),
	[DIV6_USB] = SH_CLK_DIV6_EXT(NULL, USBCKCR, 8, 0,
			usbckcr_parent, ARRAY_SIZE(usbckcr_parent), 6, 1),
	[DIV6_HDMI] = SH_CLK_DIV6_EXT(NULL, HDMICKCR, 8, 0,
			hdmickcr_parent, ARRAY_SIZE(hdmickcr_parent), 6, 2),
};

enum {	MSTP031, MSTP030, MSTP029,
	MSTP025,
	MSTP023, MSTP022, MSTP021,
	MSTP019, MSTP017, MSTP016, MSTP015,
	MSTP007,
	MSTP004, MSTP003,
	MSTP000,

	MSTP130, MSTP129, MSTP128, MSTP127,
	MSTP125, MSTP124,
	MSTP117, MSTP116, MSTP115,
	MSTP113, MSTP112, MSTP111,
	MSTP107, MSTP106,
	MSTP104, MSTP103, MSTP102, MSTP101, MSTP100,

	MSTP230, MSTP229,
	MSTP226, MSTP225,
	MSTP223, MSTP222,
	MSTP218, MSTP217, MSTP216, MSTP215,
	MSTP214, MSTP213, MSTP212,
	MSTP210, MSTP209, MSTP208, MSTP207, MSTP206, MSTP205,
	MSTP204, MSTP203, MSTP202, MSTP201, MSTP200,

	MSTP331, MSTP330, MSTP329, MSTP328, MSTP327,
	MSTP325, MSTP324, MSTP323,
	MSTP320,
	MSTP316, MSTP315, MSTP314, MSTP313, MSTP312, MSTP311,
	MSTP310, MSTP309, MSTP308,
	MSTP304,

	MSTP428,
	MSTP422, MSTP421,
	MSTP416, MSTP415,
	MSTP409, MSTP408, MSTP407, MSTP406, MSTP405,
	MSTP404, MSTP403, MSTP402,
	MSTP400,

	MSTP_NR
};

#define MSTP(_parent, _reg, _bit, _flags) \
	SH_CLK_MSTP32_EXT(_parent, SMSTPCR##_reg, MSTPSR##_reg, _bit, _flags)

static struct clk mstp_clks[MSTP_NR] = {
	[MSTP031] /*RT TLB*/	= MSTP(&div4_clks[DIV4_I], 0, 31, 0),
	[MSTP030] /*RT IC*/	= MSTP(&div4_clks[DIV4_I], 0, 30, 0),
	[MSTP029] /*RT OC*/	= MSTP(&div4_clks[DIV4_I], 0, 29, 0),
	[MSTP025] /*RT FPU*/	= MSTP(&div4_clks[DIV4_I], 0, 25, 0),
	[MSTP023] /*RT_DMAC2*/	= MSTP(&div4_clks[DIV4_S], 0, 23, 0),
	[MSTP022] /*INTCS*/	= MSTP(&div4_clks[DIV4_HP], 0, 22, 0),
	[MSTP021] /*RT-DMAC1*/	= MSTP(&div4_clks[DIV4_S], 0, 21, 0),
	[MSTP019] /*H-UDI*/	= MSTP(&div4_clks[DIV4_S], 0, 19, 0),
	[MSTP017] /*UBC*/	= MSTP(&div4_clks[DIV4_S], 0, 17, 0),
	[MSTP015] /*ILRAM*/	= MSTP(&div4_clks[DIV4_S], 0, 15, 0),
	[MSTP007] /*ICB*/	= MSTP(&div4_clks[DIV4_S], 0, 7, 0),
	[MSTP004] /*A2S*/	= MSTP(&div4_clks[DIV4_S], 0, 4, 0),
	[MSTP003] /*S2A*/	= MSTP(&div4_clks[DIV4_S], 0, 3, 0),
	[MSTP000] /*MSIOF0*/	= MSTP(&div6_reparent_clks[DIV6_SUB], 0, 0, 0),

	[MSTP130] /*2DG*/	= MSTP(&div4_clks[DIV4_S], 1, 30, 0),
	[MSTP129] /*SSP*/	= MSTP(&div4_clks[DIV4_S], 1, 29, 0),
	[MSTP128] /*CEU21*/	= MSTP(&div4_clks[DIV4_S], 1, 28, 0),
	[MSTP127] /*CEU20*/	= MSTP(&div4_clks[DIV4_S], 1, 27, 0),
	[MSTP125] /*TMU0*/	= MSTP(&div6_reparent_clks[DIV6_SUB], 1, 25, 0),
	[MSTP124] /*CMT0*/	= MSTP(&r_clk, 1, 24, 0),
	[MSTP117] /*LCDC1*/	= MSTP(&div4_clks[DIV4_B], 1, 17, 0),
	[MSTP116] /*IIC0*/	= MSTP(&div6_reparent_clks[DIV6_SUB], 1, 16, 0),
	[MSTP115] /*2D-DMAC*/	= MSTP(&div4_clks[DIV4_B], 1, 15, 0),
	[MSTP113] /*MERAM*/	= MSTP(&div4_clks[DIV4_S], 1, 13, 0),
	[MSTP112] /*SGX*/	= MSTP(&div4_clks[DIV4_ZG], 1, 12, 0),
	[MSTP111] /*TMU1*/	= MSTP(&div6_reparent_clks[DIV6_SUB], 1, 11, 0),
	[MSTP107] /*VPC*/	= MSTP(&div4_clks[DIV4_B], 1, 7, 0),
	[MSTP106] /*JPU*/	= MSTP(&div4_clks[DIV4_B], 1, 6, 0),
	[MSTP104] /*DISP*/	= MSTP(&div4_clks[DIV4_B], 1, 4, 0),
	[MSTP103] /*VIO6C*/	= MSTP(&div4_clks[DIV4_S], 1, 3, 0),
	[MSTP102] /*VPU5F*/	= MSTP(&div4_clks[DIV4_M1], 1, 2, 0),
	[MSTP101] /*VCP1*/	= MSTP(&div4_clks[DIV4_M1], 1, 1, 0),
	[MSTP100] /*LCDC0*/	= MSTP(&div4_clks[DIV4_B], 1, 0, 0),

	[MSTP230] /*SCIFA6*/	= MSTP(&div6_reparent_clks[DIV6_SUB], 2, 30, 0),
	[MSTP229] /*INTCA*/	= MSTP(&div4_clks[DIV4_HP], 2, 29, 0),
	[MSTP226] /*BBIF1*/	= MSTP(&div6_reparent_clks[DIV6_SUB], 2, 26, 0),
	[MSTP225] /*BBIF2*/	= MSTP(&div6_reparent_clks[DIV6_SUB2], 2, 25, 0),
	[MSTP223] /*SPU*/	= MSTP(&div6_reparent_clks[DIV6_SUB], 2, 23, 0),
	[MSTP222] /*SCIFA7*/	= MSTP(&div6_reparent_clks[DIV6_SUB], 2, 22, 0),
	[MSTP218] /*SY-DMAC1*/	= MSTP(&div4_clks[DIV4_S], 2, 18, 0),
	[MSTP217] /*SY-DMAC2*/	= MSTP(&div4_clks[DIV4_S], 2, 17, 0),
	[MSTP216] /*SY-DMAC3*/	= MSTP(&div4_clks[DIV4_S], 2, 16, 0),
	[MSTP215] /*DDM*/	= MSTP(&div4_clks[DIV4_S], 2, 15, 0),
	[MSTP214] /*USB-DMAC*/	= MSTP(&div4_clks[DIV4_HP], 2, 14, 0),
	[MSTP213] /*MFIS*/	= MSTP(&div4_clks[DIV4_HP], 2, 13, 0),
	[MSTP212] /*MFIM*/	= MSTP(&div4_clks[DIV4_HP], 2, 12, 0),
	[MSTP210] /*MMFROM*/	= MSTP(NULL, 2, 10, 0),
	[MSTP209] /*SMFRAM*/	= MSTP(NULL, 2, 9, 0),
	[MSTP208] /*MSIOF1*/	= MSTP(&div6_reparent_clks[DIV6_SUB], 2, 8, 0),
	[MSTP207] /*SCIFA5*/	= MSTP(&div6_reparent_clks[DIV6_SUB], 2, 7, 0),
	[MSTP206] /*SCIFB*/	= MSTP(&div6_reparent_clks[DIV6_SUB], 2, 6, 0),
	[MSTP205] /*MSIOF2*/	= MSTP(&div6_reparent_clks[DIV6_SUB], 2, 5, 0),
	[MSTP204] /*SCIFA0*/	= MSTP(&div6_reparent_clks[DIV6_SUB], 2, 4, 0),
	[MSTP203] /*SCIFA1*/	= MSTP(&div6_reparent_clks[DIV6_SUB], 2, 3, 0),
	[MSTP202] /*SCIFA2*/	= MSTP(&div6_reparent_clks[DIV6_SUB], 2, 2, 0),
	[MSTP201] /*SCIFA3*/	= MSTP(&div6_reparent_clks[DIV6_SUB], 2, 1, 0),
	[MSTP200] /*SCIFA4*/	= MSTP(&div6_reparent_clks[DIV6_SUB], 2, 0, 0),

	[MSTP331] /*VPU2*/	= MSTP(&div4_clks[DIV4_HP], 3, 31, 0),
	[MSTP330] /*ATAPI*/	= MSTP(&div6_reparent_clks[DIV6_SUB], 3, 30, 0),
	[MSTP329] /*CMT1*/	= MSTP(&r_clk, 3, 29, 0),
	[MSTP328] /*FSI*/	= MSTP(&div4_clks[DIV4_HP], 3, 28, 0),
	[MSTP327] /*FMSI*/	= MSTP(&div6_reparent_clks[DIV6_SUB2], 3, 27, 0),
	[MSTP325] /*IrDA*/	= MSTP(&div6_reparent_clks[DIV6_SUB], 3, 25, 0),
	[MSTP324] /*IrREM*/	= MSTP(&div6_reparent_clks[DIV6_SUB], 3, 24, 0),
	[MSTP323] /*IIC1*/	= MSTP(&div6_reparent_clks[DIV6_SUB], 3, 23, 0),
	[MSTP320] /*USBF*/	= MSTP(&div4_clks[DIV4_HP], 3, 20, 0),
	[MSTP316] /*SHWYSTAT*/	= MSTP(&div4_clks[DIV4_S], 3, 16, 0),
	[MSTP315] /*FLCTL*/	= MSTP(&div4_clks[DIV4_HP], 3, 15, 0),
	[MSTP314] /*SDHI0*/	= MSTP(&div4_clks[DIV4_HP], 3, 14, 0),
	[MSTP313] /*SDHI1*/	= MSTP(&div4_clks[DIV4_HP], 3, 13, 0),
	[MSTP312] /*MMC*/	= MSTP(&div4_clks[DIV4_HP], 3, 12, 0),
	[MSTP311] /*RSPI*/	= MSTP(&div6_reparent_clks[DIV6_SUB], 3, 11, 0),
	[MSTP310] /*SIM*/	= MSTP(&div4_clks[DIV4_HP], 3, 10, 0),
	[MSTP309] /*GEther*/	= MSTP(&div4_clks[DIV4_HP], 3, 9, 0),
	[MSTP308] /*MFIM*/	= MSTP(&div4_clks[DIV4_HP], 3, 8, 0),
	[MSTP304] /*TPU0*/	= MSTP(&div4_clks[DIV4_CP], 3, 4, 0),

	[MSTP428] /*SDENC*/	= MSTP(&div4_clks[DIV4_B], 4, 28, 0),
	[MSTP422] /*ACMFI1*/	= MSTP(&div4_clks[DIV4_HP], 4, 22, 0),
	[MSTP421] /*ACMFI2*/	= MSTP(&div4_clks[DIV4_HP], 4, 21, 0),
	[MSTP416] /*USBHOST*/	= MSTP(&div4_clks[DIV4_HP], 4, 16, 0),
	[MSTP415] /*SDHI2*/	= MSTP(&div4_clks[DIV4_HP], 4, 15, 0),
	[MSTP409] /*ACMFIRAM*/	= MSTP(NULL, 4, 9, 0),
	[MSTP408] /*SIB*/	= MSTP(&div4_clks[DIV4_HP], 4, 8, 0),
	[MSTP407] /*USB-Func*/	= MSTP(&mstp_clks[MSTP320], 4, 7, 0),
	[MSTP406] /*USB-Phy*/	= MSTP(&div4_clks[DIV4_HP], 4, 6, 0),
	[MSTP405] /*CMT4*/	= MSTP(&r_clk, 4, 5, 0),
	[MSTP404] /*CMT3*/	= MSTP(&r_clk, 4, 4, 0),
	[MSTP403] /*KEYSC*/	= MSTP(&div4_clks[DIV4_CP], 4, 3, 0),
	[MSTP402] /*RWDT0*/	= MSTP(&r_clk, 4, 2, 0),
	[MSTP400] /*CMT2*/	= MSTP(&r_clk, 4, 0, 0),
};

#define CLKDEV_CON_ID(_id, _clk) { .con_id = _id, .clk = _clk }
#define CLKDEV_DEV_ID(_id, _clk) { .dev_id = _id, .clk = _clk }
#define CLKDEV_ICK_ID(_cid, _did, _clk) \
		{ .con_id = _cid, .dev_id = _did, .clk = _clk }

static struct clk_lookup lookups[] = {
	/* main clocks */
	CLKDEV_CON_ID("extalr", &extalr_clk),
	CLKDEV_CON_ID("extal1", &extal1_clk),
	CLKDEV_CON_ID("extal2", &extal2_clk),
	CLKDEV_CON_ID("extal1_div2", &extal1_div2_clk),
	CLKDEV_CON_ID("extal1_div1024", &extal1_div1024_clk),
	CLKDEV_CON_ID("extal2_div2", &extal2_div2_clk),
	CLKDEV_CON_ID("dv_clk", &dv_clk),
	CLKDEV_CON_ID("system_clk", &system_clk),
	CLKDEV_CON_ID("system_div2_clk", &system_div2_clk),
	CLKDEV_CON_ID("r_clk", &r_clk),
	CLKDEV_CON_ID("pllc0_clk", &pllc0_clk),
	CLKDEV_CON_ID("pllc1_clk", &pllc1_clk),
	CLKDEV_CON_ID("pllc1_div2_clk", &pllc1_div2_clk),
	CLKDEV_CON_ID("usb24s", &usb24s_clk),
	CLKDEV_CON_ID("usb24s_div2", &usb24s_div2_clk),
	CLKDEV_CON_ID("fmsick_clk", &fmsick_clk),
	CLKDEV_CON_ID("fmsock_clk", &fmsock_clk),
	CLKDEV_CON_ID("fsiack_clk", &fsiack_clk),
	CLKDEV_CON_ID("z_clk", &z_clk),

	/* DIV4 clocks */
	CLKDEV_CON_ID("i_clk", &div4_clks[DIV4_I]),
	CLKDEV_CON_ID("zg_clk", &div4_clks[DIV4_ZG]),
	CLKDEV_CON_ID("b_clk", &div4_clks[DIV4_B]),
	CLKDEV_CON_ID("m1_clk", &div4_clks[DIV4_M1]),
	CLKDEV_CON_ID("hp_clk", &div4_clks[DIV4_HP]),
	CLKDEV_CON_ID("hpp_clk", &div4_clks[DIV4_HPP]),
	CLKDEV_CON_ID("usbp_clk", &div4_clks[DIV4_USBP]),
	CLKDEV_CON_ID("s_clk", &div4_clks[DIV4_S]),
	CLKDEV_CON_ID("zb_clk", &div4_clks[DIV4_ZB]),
	CLKDEV_CON_ID("m3_clk", &div4_clks[DIV4_M3]),
	CLKDEV_CON_ID("cp_clk", &div4_clks[DIV4_CP]),

	/* DIV6 clocks */
	CLKDEV_CON_ID("vck1_clk", &div6_clks[DIV6_VCK1]),
	CLKDEV_CON_ID("vck2_clk", &div6_clks[DIV6_VCK2]),
	CLKDEV_CON_ID("spu_clk", &div6_clks[DIV6_SPU]),
	CLKDEV_CON_ID("stpr_clk", &div6_clks[DIV6_STPR]),

	CLKDEV_CON_ID("sub_clk", &div6_reparent_clks[DIV6_SUB]),
	CLKDEV_CON_ID("sub2_clk", &div6_reparent_clks[DIV6_SUB2]),
	CLKDEV_CON_ID("vou_clk", &div6_reparent_clks[DIV6_VOU]),
	CLKDEV_CON_ID("fmsi_clk", &div6_reparent_clks[DIV6_FMSI]),
	CLKDEV_CON_ID("fmso_clk", &div6_reparent_clks[DIV6_FMSO]),
	CLKDEV_CON_ID("fsia_clk", &div6_reparent_clks[DIV6_FSIA]),
	CLKDEV_CON_ID("usb_clk", &div6_reparent_clks[DIV6_USB]),
	CLKDEV_CON_ID("hdmi_clk", &div6_reparent_clks[DIV6_HDMI]),

	/* MSTP32 clocks */
	CLKDEV_CON_ID("usb0_dmac", &mstp_clks[MSTP214]),
	CLKDEV_CON_ID("usbh", &mstp_clks[MSTP416]),
	CLKDEV_CON_ID("usb_phy", &mstp_clks[MSTP406]),
	CLKDEV_CON_ID("keysc0", &mstp_clks[MSTP403]),
	CLKDEV_CON_ID("mmcif0", &mstp_clks[MSTP312]),
	CLKDEV_CON_ID("msiof0", &mstp_clks[MSTP000]),
	CLKDEV_CON_ID("msiof1", &mstp_clks[MSTP208]),
	CLKDEV_CON_ID("msiof2", &mstp_clks[MSTP205]),
	CLKDEV_CON_ID("tmu0", &mstp_clks[MSTP125]),
	CLKDEV_CON_ID("tmu1", &mstp_clks[MSTP111]),
	CLKDEV_CON_ID("mfis", &mstp_clks[MSTP213]),
	CLKDEV_CON_ID("mfim", &mstp_clks[MSTP212]),
	CLKDEV_CON_ID("irda", &mstp_clks[MSTP325]),
	CLKDEV_CON_ID("irrem", &mstp_clks[MSTP324]),
	CLKDEV_CON_ID("flctl", &mstp_clks[MSTP315]),
	CLKDEV_CON_ID("rwdt0", &mstp_clks[MSTP402]),
	CLKDEV_CON_ID("atapi", &mstp_clks[MSTP330]),
	CLKDEV_CON_ID("fsi", &mstp_clks[MSTP328]),
	CLKDEV_CON_ID("fmsi", &mstp_clks[MSTP327]),
	CLKDEV_CON_ID("sgx", &mstp_clks[MSTP112]),
	CLKDEV_CON_ID("ceu0", &mstp_clks[MSTP127]),
	CLKDEV_CON_ID("ceu1", &mstp_clks[MSTP128]),
	CLKDEV_CON_ID("tpu", &mstp_clks[MSTP304]),

	CLKDEV_DEV_ID("r8a66597_udc.0", &mstp_clks[MSTP407]),
	CLKDEV_DEV_ID("sh_mobile_sdhi.0", &mstp_clks[MSTP314]),
	CLKDEV_DEV_ID("sh_mobile_sdhi.1", &mstp_clks[MSTP313]),
	CLKDEV_DEV_ID("sh_mobile_sdhi.2", &mstp_clks[MSTP415]),
	CLKDEV_DEV_ID("sh_mmcif.0", &mstp_clks[MSTP312]),
	CLKDEV_DEV_ID("sh-eth.0", &mstp_clks[MSTP309]),
	CLKDEV_DEV_ID("sh-sci.0", &mstp_clks[MSTP204]),
	CLKDEV_DEV_ID("sh-sci.1", &mstp_clks[MSTP203]),
	CLKDEV_DEV_ID("sh-sci.2", &mstp_clks[MSTP202]),
	CLKDEV_DEV_ID("sh-sci.3", &mstp_clks[MSTP201]),
	CLKDEV_DEV_ID("sh-sci.4", &mstp_clks[MSTP200]),
	CLKDEV_DEV_ID("sh-sci.5", &mstp_clks[MSTP207]),
	CLKDEV_DEV_ID("sh-sci.6", &mstp_clks[MSTP230]),
	CLKDEV_DEV_ID("sh-sci.7", &mstp_clks[MSTP222]),
	CLKDEV_DEV_ID("sh-sci.8", &mstp_clks[MSTP206]),
	CLKDEV_DEV_ID("sh_cmt.10", &mstp_clks[MSTP329]),
	CLKDEV_DEV_ID("sh_cmt.14", &mstp_clks[MSTP329]),
	CLKDEV_DEV_ID("sh_tmu.0", &mstp_clks[MSTP125]),
	CLKDEV_DEV_ID("sh_tmu.1", &mstp_clks[MSTP111]),
	CLKDEV_DEV_ID("sh_keysc.0", &mstp_clks[MSTP403]),
	CLKDEV_DEV_ID("i2c-sh_mobile.0", &mstp_clks[MSTP116]),
	CLKDEV_DEV_ID("i2c-sh_mobile.1", &mstp_clks[MSTP323]),
	CLKDEV_DEV_ID("spi_sh_msiof.0", &mstp_clks[MSTP000]),
	CLKDEV_DEV_ID("spi_sh_msiof.1", &mstp_clks[MSTP208]),
	CLKDEV_DEV_ID("spi_sh_msiof.2", &mstp_clks[MSTP205]),
	CLKDEV_DEV_ID("sh_mobile_lcdc_fb.0", &mstp_clks[MSTP100]),
	CLKDEV_DEV_ID("sh_mobile_lcdc_fb.1", &mstp_clks[MSTP117]),
	CLKDEV_DEV_ID("sh-dma-engine.0", &mstp_clks[MSTP218]),
	CLKDEV_DEV_ID("sh-dma-engine.1", &mstp_clks[MSTP217]),
	CLKDEV_DEV_ID("sh-dma-engine.2", &mstp_clks[MSTP216]),
	CLKDEV_DEV_ID("sh_keysc.0", &mstp_clks[MSTP403]),
	CLKDEV_DEV_ID("sh_mobile_ceu.0", &mstp_clks[MSTP127]),
	CLKDEV_DEV_ID("sh_mobile_ceu.1", &mstp_clks[MSTP128]),

	CLKDEV_ICK_ID("icka", "sh_fsi2", &div6_reparent_clks[DIV6_FSIA]),
};

void __init sh7740_clock_init(int md0, int md1, int md2)
{
	int k, ret = 0;

	/* detect system clock parent */
	if (md1)
		system_clk.parent = &extal1_div2_clk;
	else
		system_clk.parent = &extal1_clk;

	/* detect RCLK parent */
	if (md2)
		r_clk.parent = &extal1_div1024_clk;
	else
		r_clk.parent = &extalr_clk;

	/* detect z_clk */
	z_clk.priv = (void *)(((__raw_readl(FRQCRB) >> 24) & 0x1f) + 1);

	for (k = 0; !ret && (k < ARRAY_SIZE(main_clks)); k++)
		ret = clk_register(main_clks[k]);

	if (!ret)
		ret = sh_clk_div4_register(div4_clks, DIV4_NR, &div4_table);

	if (!ret)
		ret = sh_clk_div6_register(div6_clks, DIV6_NR);

	if (!ret)
		ret = sh_clk_div6_reparent_register(
				div6_reparent_clks, DIV6_REPARENT_NR);

	if (!ret)
		ret = sh_clk_mstp32_register(mstp_clks, MSTP_NR);

	clkdev_add_table(lookups, ARRAY_SIZE(lookups));

	if (!ret)
		clk_init();
	else
		panic("failed to setup upd77660 clocks\n");
}
