/*
 * Driver for IMX036 CMOS Image Sensor from Sony
 *
 * Copyright (C) 2011, Atmark Techno, Inc.
 *
 * Which was Based on imx074.c
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/delay.h>
#include <linux/i2c.h>
#include <linux/slab.h>
#include <linux/videodev2.h>

#include <media/soc_camera.h>
#include <media/soc_mediabus.h>
#include <media/v4l2-subdev.h>
#include <media/v4l2-chip-ident.h>

/* IMX036 registers */
/* todo */

/* IMX036 supported geometry */
#define IMX036_WIDTH			1920
#define IMX036_HEIGHT			1080

/* IMX036 has only one fixed colorspace per pixelcode */
struct imx036_datafmt {
	enum v4l2_mbus_pixelcode	code;
	enum v4l2_colorspace		colorspace;
};

struct imx036 {
	struct v4l2_subdev		subdev;
	const struct imx036_datafmt	*fmt;
};

static const struct imx036_datafmt imx036_colour_fmts[] = {
	{
		.code		= V4L2_MBUS_FMT_UYVY8_2X8,
		.colorspace	= V4L2_COLORSPACE_JPEG,
	},
};

static struct imx036 *to_imx036(const struct i2c_client *client)
{
	return container_of(i2c_get_clientdata(client), struct imx036, subdev);
}

/* Find a data format by a pixel code in an array */
static const struct imx036_datafmt *imx036_find_datafmt(enum v4l2_mbus_pixelcode code)
{
	int i;

	for (i = 0; i < ARRAY_SIZE(imx036_colour_fmts); i++)
		if (imx036_colour_fmts[i].code == code)
			return imx036_colour_fmts + i;

	return NULL;
}

static int reg_write(struct i2c_client *client, const u16 addr, const u8 data)
{
	struct i2c_adapter *adap = client->adapter;
	struct i2c_msg msg;
	unsigned char tx[3];
	int ret;

	msg.addr = client->addr;
	msg.buf = tx;
	msg.len = 3;
	msg.flags = 0;

	tx[0] = addr >> 8;
	tx[1] = addr & 0xff;
	tx[2] = data;

	ret = i2c_transfer(adap, &msg, 1);

	mdelay(2);

	return ret == 1 ? 0 : -EIO;
}

static int reg_read(struct i2c_client *client, const u16 addr)
{
	u8 buf[2] = {addr >> 8, addr & 0xff};
	int ret;
	struct i2c_msg msgs[] = {
		{
			.addr  = client->addr,
			.flags = 0,
			.len   = 2,
			.buf   = buf,
		}, {
			.addr  = client->addr,
			.flags = I2C_M_RD,
			.len   = 2,
			.buf   = buf,
		},
	};

	ret = i2c_transfer(client->adapter, msgs, ARRAY_SIZE(msgs));
	if (ret < 0) {
		dev_warn(&client->dev, "Reading register %x from %x failed\n",
			 addr, client->addr);
		return ret;
	}

	return buf[0] & 0xff; /* no sign-extension */
}

static int imx036_try_fmt(struct v4l2_subdev *sd,
			  struct v4l2_mbus_framefmt *mf)
{
	const struct imx036_datafmt *fmt = imx036_find_datafmt(mf->code);

	dev_dbg(sd->v4l2_dev->dev, "%s(%u)\n", __func__, mf->code);

	if (!fmt) {
		mf->code	= imx036_colour_fmts[0].code;
		mf->colorspace	= imx036_colour_fmts[0].colorspace;
	}

	mf->width	= IMX036_WIDTH;
	mf->height	= IMX036_HEIGHT;
	mf->field	= V4L2_FIELD_NONE;

	return 0;
}

static int imx036_s_fmt(struct v4l2_subdev *sd,
			struct v4l2_mbus_framefmt *mf)
{
	struct i2c_client *client = v4l2_get_subdevdata(sd);
	struct imx036 *priv = to_imx036(client);

	dev_dbg(sd->v4l2_dev->dev, "%s(%u)\n", __func__, mf->code);

	/* MIPI CSI could have changed the format, double-check */
	if (!imx036_find_datafmt(mf->code))
		return -EINVAL;

	imx036_try_fmt(sd, mf);

	priv->fmt = imx036_find_datafmt(mf->code);

	return 0;
}

static int imx036_g_fmt(struct v4l2_subdev *sd,
			struct v4l2_mbus_framefmt *mf)
{
	struct i2c_client *client = v4l2_get_subdevdata(sd);
	struct imx036 *priv = to_imx036(client);

	const struct imx036_datafmt *fmt = priv->fmt;

	mf->code	= fmt->code;
	mf->colorspace	= fmt->colorspace;
	mf->width	= IMX036_WIDTH;
	mf->height	= IMX036_HEIGHT;
	mf->field	= V4L2_FIELD_NONE;

	return 0;
}

static int imx036_g_crop(struct v4l2_subdev *sd, struct v4l2_crop *a)
{
	struct v4l2_rect *rect = &a->c;

	a->type		= V4L2_BUF_TYPE_VIDEO_CAPTURE;
	rect->top	= 0;
	rect->left	= 0;
	rect->width	= IMX036_WIDTH;
	rect->height	= IMX036_HEIGHT;

	return 0;
}

static int imx036_s_crop(struct v4l2_subdev *sd, struct v4l2_crop *a)
{
	return 0;
}

static int imx036_cropcap(struct v4l2_subdev *sd, struct v4l2_cropcap *a)
{
	a->bounds.left			= 0;
	a->bounds.top			= 0;
	a->bounds.width			= IMX036_WIDTH;
	a->bounds.height		= IMX036_HEIGHT;
	a->defrect			= a->bounds;
	a->type				= V4L2_BUF_TYPE_VIDEO_CAPTURE;
	a->pixelaspect.numerator	= 1;
	a->pixelaspect.denominator	= 1;

	return 0;
}

static int imx036_enum_fmt(struct v4l2_subdev *sd, unsigned int index,
			   enum v4l2_mbus_pixelcode *code)
{
	if ((unsigned int)index >= ARRAY_SIZE(imx036_colour_fmts))
		return -EINVAL;

	*code = imx036_colour_fmts[index].code;
	return 0;
}

static int imx036_s_stream(struct v4l2_subdev *sd, int enable)
{
	struct i2c_client *client = v4l2_get_subdevdata(sd);

	/* MODE_SELECT: stream or standby */
	return 0;//reg_write(client, MODE_SELECT, !!enable);
}

static int imx036_g_chip_ident(struct v4l2_subdev *sd,
			       struct v4l2_dbg_chip_ident *id)
{
	struct i2c_client *client = v4l2_get_subdevdata(sd);

	if (id->match.type != V4L2_CHIP_MATCH_I2C_ADDR)
		return -EINVAL;

	if (id->match.addr != client->addr)
		return -ENODEV;

	id->ident	= V4L2_IDENT_IMX036;
	id->revision	= 0;

	return 0;
}

static struct v4l2_subdev_video_ops imx036_subdev_video_ops = {
	.s_stream	= imx036_s_stream,
	.s_mbus_fmt	= imx036_s_fmt,
	.g_mbus_fmt	= imx036_g_fmt,
	.try_mbus_fmt	= imx036_try_fmt,
	.enum_mbus_fmt	= imx036_enum_fmt,
	.g_crop		= imx036_g_crop,
	.s_crop		= imx036_s_crop,
	.cropcap	= imx036_cropcap,
};

static struct v4l2_subdev_core_ops imx036_subdev_core_ops = {
	.g_chip_ident	= imx036_g_chip_ident,
};

static struct v4l2_subdev_ops imx036_subdev_ops = {
	.core	= &imx036_subdev_core_ops,
	.video	= &imx036_subdev_video_ops,
};

/*
 * We have to provide soc-camera operations, but we don't have anything to say
 * there. The MIPI CSI2 driver will provide .query_bus_param and .set_bus_param
 */
static unsigned long imx036_query_bus_param(struct soc_camera_device *icd)
{
	struct soc_camera_link *icl = to_soc_camera_link(icd);
	unsigned long flags = SOCAM_MASTER | SOCAM_VSYNC_ACTIVE_HIGH |
		SOCAM_HSYNC_ACTIVE_HIGH | SOCAM_DATA_ACTIVE_HIGH;
	flags |= SOCAM_PCLK_SAMPLE_RISING;
	flags |= SOCAM_DATAWIDTH_16;
	return soc_camera_apply_sensor_flags(icl, flags);
}

static int imx036_set_bus_param(struct soc_camera_device *icd,
				 unsigned long flags)
{
	return 0;
}

static struct soc_camera_ops imx036_ops = {
	.query_bus_param	= imx036_query_bus_param,
	.set_bus_param		= imx036_set_bus_param,
};

static int imx036_video_probe(struct soc_camera_device *icd,
			      struct i2c_client *client)
{
	/* todo */
	dev_info(&client->dev, "%s chip ID %04x\n", "imx036", 0);
	return 0;
}

static int imx036_probe(struct i2c_client *client,
			const struct i2c_device_id *did)
{
	struct imx036 *priv;
	struct soc_camera_device *icd = client->dev.platform_data;
	struct i2c_adapter *adapter = to_i2c_adapter(client->dev.parent);
	struct soc_camera_link *icl;
	int ret;

	if (!icd) {
		dev_err(&client->dev, "IMX036: missing soc-camera data!\n");
		return -EINVAL;
	}

	icl = to_soc_camera_link(icd);
	if (!icl) {
		dev_err(&client->dev, "IMX036: missing platform data!\n");
		return -EINVAL;
	}

	if (!i2c_check_functionality(adapter, I2C_FUNC_SMBUS_BYTE_DATA)) {
		dev_warn(&adapter->dev,
			 "I2C-Adapter doesn't support I2C_FUNC_SMBUS_BYTE\n");
		return -EIO;
	}

	priv = kzalloc(sizeof(struct imx036), GFP_KERNEL);
	if (!priv)
		return -ENOMEM;

	v4l2_i2c_subdev_init(&priv->subdev, client, &imx036_subdev_ops);

	icd->ops	= &imx036_ops;
	priv->fmt	= &imx036_colour_fmts[0];

	ret = imx036_video_probe(icd, client);
	if (ret < 0) {
		icd->ops = NULL;
		kfree(priv);
		return ret;
	}

	return ret;
}

static int imx036_remove(struct i2c_client *client)
{
	struct imx036 *priv = to_imx036(client);
	struct soc_camera_device *icd = client->dev.platform_data;
	struct soc_camera_link *icl = to_soc_camera_link(icd);

	icd->ops = NULL;
	if (icl->free_bus)
		icl->free_bus(icl);
	kfree(priv);

	return 0;
}

static const struct i2c_device_id imx036_id[] = {
	{ "imx036", 0 },
	{ }
};
MODULE_DEVICE_TABLE(i2c, imx036_id);

static struct i2c_driver imx036_i2c_driver = {
	.driver = {
		.name = "imx036",
	},
	.probe		= imx036_probe,
	.remove		= imx036_remove,
	.id_table	= imx036_id,
};

static int __init imx036_mod_init(void)
{
	return i2c_add_driver(&imx036_i2c_driver);
}

static void __exit imx036_mod_exit(void)
{
	i2c_del_driver(&imx036_i2c_driver);
}

module_init(imx036_mod_init);
module_exit(imx036_mod_exit);

MODULE_DESCRIPTION("Sony IMX036 Camera driver");
MODULE_AUTHOR("Atmark Techno, Inc.");
MODULE_LICENSE("GPL v2");
