/*
 *  OHCI HCD (Host Controller Driver) for SHMOBILE
 *
 *  Copyright (C) 2011 Renesas Electronics Corporation
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2
 *  as published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
 */

#include <linux/clk.h>
#include <linux/device.h>
#include <linux/signal.h>
#include <linux/platform_device.h>
#include <linux/miscdevice.h>
#include <linux/usb/rmobile.h>
#include <linux/pm_runtime.h>

struct rmobile_ohci {
	/* must be 1st member here for hcd_to_ohci() to work */
	struct ohci_hcd ohci;

	u32		phys_base;
	void __iomem	*base;

	int		vbus_flag;

	/* platform data */
	struct rmobile_usb_platdata *pdata;

	/* backup area */
	struct rmobile_ohci_reg ohci_reg;
};

#define to_rmobile_ohci(hcd)	(struct rmobile_ohci *)hcd_to_ohci(hcd)

/*-------------------------------------------------------------------------*/
static void rmobile_vbus_control(struct rmobile_ohci *ohci, int on)
{
	if (ohci->pdata->set_vbus)
		ohci->pdata->set_vbus(on);
	ohci->vbus_flag = on ? 0 : 1;
}

static int rmobile_over_current_check(struct rmobile_ohci *ohci)
{
	if (ohci->pdata->check_oci()) {
		rmobile_vbus_control(ohci, 0);
		err("***** USB Over Current !!\n");
		return 1;
	}
	return 0;
}

/*-------------------------------------------------------------------------*/

static void rmobile_hc_start(struct usb_hcd *hcd)
{
	u32 data, tmp;
	struct clk *clk;
	struct rmobile_ohci *ohci = to_rmobile_ohci(hcd);

	if (ohci->pdata->common_clock)
		ohci->pdata->common_clock(1);
	else {
		/* Start Clock */
		clk = clk_get(NULL, "usbh");
		if (!IS_ERR(clk)) {
			clk_enable(clk);
			clk_put(clk);
		}
		clk = clk_get(NULL, "usbp");
		if (!IS_ERR(clk)) {
			clk_enable(clk);
			clk_put(clk);
		}

		/* Clock & Reset & Direct Power Down */
		data = __raw_readl(ohci->base + USBCTR);
		data &= ~(DIRPD | PCICLK_MASK | USBH_RST);
		__raw_writel(data, ohci->base + USBCTR);
		__raw_writel(data & ~PLL_RST, ohci->base + USBCTR);
	}

	data = SMODE_READY_CTR | MMODE_HBUSREQ |
			MMODE_WR_INCR | MMODE_BYTE_BURST | MMODE_HTRANS;

	tmp = __raw_readl(ohci->base + AHB_BUS_CTR);
	if (data == tmp)
		return;

	/****** AHB-PCI Bridge Communication Registers ******/
	__raw_writel(data, ohci->base + AHB_BUS_CTR);
	__raw_writel((PHYS_OFFSET & 0xf0000000) | PREFETCH,
			ohci->base + PCIAHB_WIN1_CTR);
	__raw_writel(0xf0000000 | PREFETCH, ohci->base + PCIAHB_WIN2_CTR);
	__raw_writel(ohci->phys_base | PCIWIN2_PCICMD,
			ohci->base + AHBPCI_WIN2_CTR);
	data = __raw_readl(ohci->base + PCI_ARBITER_CTR);
	__raw_writel(data | PCIBP_MODE | PCIREQ1 | PCIREQ0,
			ohci->base + PCI_ARBITER_CTR);

	/****** PCI Configuration Registers for AHBPCI ******/
	__raw_writel(PCIWIN1_PCICMD | AHB_CFG_AHBPCI,
			ohci->base + AHBPCI_WIN1_CTR);
	__raw_writel(ohci->phys_base + AHBPCI_OFFSET, ohci->base + BASEAD);
	__raw_writel(PHYS_OFFSET & 0xf0000000, ohci->base + WIN1_BASEAD);
	__raw_writel(0xf0000000, ohci->base + WIN2_BASEAD);
	__raw_writel(SERREN | PERREN | MASTEREN | MEMEN,
			ohci->base + CMND_STS);

	/****** PCI Configuration Registers for OHCI/EHCI ******/
	data = PCIWIN1_PCICMD | AHB_CFG_HOST;
	__raw_writel(data, ohci->base + AHBPCI_WIN1_CTR);
	__raw_writel(ohci->phys_base + OHCI_OFFSET, ohci->base + OHCI_BASEAD);
	__raw_writel(ohci->phys_base + EHCI_OFFSET, ohci->base + EHCI_BASEAD);
	data = SERREN | PERREN | MASTEREN | MEMEN;
	__raw_writel(data, ohci->base + OHCI_CMND_STS);
	__raw_writel(data, ohci->base + EHCI_CMND_STS);

	/* PCI_INT_ENABLE */
	data = __raw_readl(ohci->base + PCI_INT_ENABLE);
	data |= USBH_PMEEN | USBH_INTBEN | USBH_INTAEN;
	__raw_writel(data | USBH_PMEEN | USBH_INTBEN | USBH_INTAEN,
			ohci->base + PCI_INT_ENABLE);
}

static void rmobile_hc_stop(struct rmobile_ohci *ohci)
{
	u32 data;
	struct clk *clk;

	rmobile_vbus_control(ohci, 0);

	if (ohci->pdata->common_clock)
		ohci->pdata->common_clock(0);
	else {
		/* Clock & Reset & Direct Power Down */
		data = __raw_readl(ohci->base + USBCTR);
		data |= DIRPD | PLL_RST | PCICLK_MASK | USBH_RST;
		__raw_writel(data, ohci->base + USBCTR);

		/* Stop Clock */
		clk = clk_get(NULL, "usbh");
		if (!IS_ERR(clk)) {
			clk_disable(clk);
			clk_put(clk);
		}
		clk = clk_get(NULL, "usbp");
		if (!IS_ERR(clk)) {
			clk_disable(clk);
			clk_put(clk);
		}
	}
}

/*-------------------------------------------------------------------------*/

static int rmobile_hub_control(
	struct usb_hcd	*hcd,
	u16		typeReq,
	u16		wValue,
	u16		wIndex,
	char		*buf,
	u16		wLength
)
{
	int retval;
	struct rmobile_ohci *ohci = to_rmobile_ohci(hcd);

	retval = ohci_hub_control(hcd, typeReq, wValue, wIndex, buf, wLength);

	/* VBUS ON/OFF control */
	if (retval == 0) {
		if (wValue == USB_PORT_FEAT_POWER) {
			if (typeReq == SetPortFeature) {
				rmobile_vbus_control(ohci, 1);
				if (ohci->pdata->oci_irq) {
					mdelay(OVER_CURRENT_TIME);
					if (!rmobile_over_current_check(ohci))
						enable_irq(ohci->pdata->oci_irq);
				}
			} else if (typeReq == ClearPortFeature) {
				if (ohci->vbus_flag) {
					if (ohci->pdata->oci_irq)
						disable_irq(ohci->pdata->oci_irq);
					rmobile_vbus_control(ohci, 0);
				} else
					ohci->vbus_flag = 1;
			}
		}
	}

	return retval;
}

static irqreturn_t rmobile_over_current_irq(int irq, void *_hcd)
{
	struct rmobile_ohci *ohci = to_rmobile_ohci(_hcd);
	rmobile_over_current_check(ohci);

	return IRQ_HANDLED;
}

/*-------------------------------------------------------------------------*/

static void rmobile_ohci_save_register(struct usb_hcd *hcd)
{
	struct rmobile_ohci_reg *p_reg;
	struct ohci_hcd *ohci = hcd_to_ohci(hcd);
	struct rmobile_ohci *rohci = to_rmobile_ohci(hcd);

	p_reg = &rohci->ohci_reg;
	if (p_reg->save_flag)
		return;

	p_reg->save_flag = 1;

	p_reg->HcInterruptEnable = ohci_readl(ohci, &ohci->regs->intrenable);
	p_reg->HcControl = ohci_readl(ohci, &ohci->regs->control);
	p_reg->HcCommandStatus = ohci_readl(ohci, &ohci->regs->cmdstatus);
	p_reg->HcHCCA = ohci_readl(ohci, &ohci->regs->hcca);
	p_reg->HcControlHeadED =
		 ohci_readl(ohci, &ohci->regs->ed_controlhead);
	p_reg->HcControlCurrentED =
		 ohci_readl(ohci, &ohci->regs->ed_controlcurrent);
	p_reg->HcBulkHeadED = ohci_readl(ohci, &ohci->regs->ed_bulkhead);
	p_reg->HcBulkCurrentED =
		 ohci_readl(ohci, &ohci->regs->ed_bulkcurrent);
	p_reg->HcFmInterval = ohci_readl(ohci, &ohci->regs->fminterval);
	p_reg->HcPeriodicStart = ohci_readl(ohci, &ohci->regs->periodicstart);
	p_reg->HcLSThreshold = ohci_readl(ohci, &ohci->regs->lsthresh);

	p_reg->HcRhDescrptorA = ohci_readl(ohci, &ohci->regs->roothub.a);
	p_reg->HcRhDescrptorB = ohci_readl(ohci, &ohci->regs->roothub.b);
	p_reg->HcRhStatus = ohci_readl(ohci, &ohci->regs->roothub.status);
	p_reg->HcRhPortStatus =
		 ohci_readl(ohci, &ohci->regs->roothub.portstatus[0]);
}

static void rmobile_ohci_load_register(struct usb_hcd *hcd)
{
	struct rmobile_ohci_reg *p_reg;
	struct ohci_hcd		*ohci = hcd_to_ohci(hcd);
	struct rmobile_ohci *rohci = to_rmobile_ohci(hcd);

	p_reg = &rohci->ohci_reg;
	if (p_reg->save_flag == 0)
		return;

	ohci_writel(ohci, p_reg->HcRhPortStatus,
		 &ohci->regs->roothub.portstatus[0]);
	ohci_writel(ohci, p_reg->HcRhStatus, &ohci->regs->roothub.status);
	ohci_writel(ohci, p_reg->HcRhDescrptorB, &ohci->regs->roothub.b);
	ohci_writel(ohci, p_reg->HcRhDescrptorA, &ohci->regs->roothub.a);

	ohci_writel(ohci, p_reg->HcLSThreshold, &ohci->regs->lsthresh);
	ohci_writel(ohci, p_reg->HcPeriodicStart, &ohci->regs->periodicstart);
	ohci_writel(ohci, p_reg->HcFmInterval, &ohci->regs->fminterval);
	ohci_writel(ohci, p_reg->HcBulkCurrentED, &ohci->regs->ed_bulkcurrent);
	ohci_writel(ohci, p_reg->HcBulkHeadED, &ohci->regs->ed_bulkhead);
	ohci_writel(ohci, p_reg->HcControlCurrentED,
		 &ohci->regs->ed_controlcurrent);
	ohci_writel(ohci, p_reg->HcControlHeadED,
		 &ohci->regs->ed_controlhead);
	ohci_writel(ohci, p_reg->HcHCCA, &ohci->regs->hcca);
	ohci_writel(ohci, p_reg->HcCommandStatus, &ohci->regs->cmdstatus);
	ohci_writel(ohci, p_reg->HcControl, &ohci->regs->control);
	ohci_writel(ohci, p_reg->HcInterruptEnable, &ohci->regs->intrenable);

	p_reg->save_flag = 0;
}

/*-------------------------------------------------------------------------*/

static int rmobile_ohci_probe(
	const struct hc_driver *driver,
	 struct platform_device *pdev
)
{
	int retval, irq;
	struct usb_hcd *hcd;
	struct rmobile_ohci *ohci;
	struct rmobile_usb_platdata *pdata;
	struct resource *r;

	pdata = pdev->dev.platform_data;
	if (!pdata)
		return -ENODEV;

	/* OHCI irq */
	irq = platform_get_irq(pdev, 0);
	if (irq < 0) {
		pr_err("no resource of IORESOURCE_IRQ");
		return -ENXIO;
	}

	hcd = usb_create_hcd(driver, &pdev->dev, "rmobile_ohci");
	if (!hcd)
		return -ENOMEM;

	r = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!r) {
		pr_err("no resource of IORESOURCE_MEM");
		retval = -ENXIO;
		goto err1;
	}

	hcd->rsrc_start = r->start + OHCI_OFFSET;
	hcd->rsrc_len = OHCI_SIZE;
	if (!request_mem_region(hcd->rsrc_start, hcd->rsrc_len, hcd_name)) {
		pr_debug("request_mem_region failed");
		retval = -EBUSY;
		goto err1;
	}

	hcd->regs = ioremap(r->start, resource_size(r));
	if (!hcd->regs) {
		pr_debug("ioremap failed");
		retval = -ENOMEM;
		goto err2;
	}

	ohci = (struct rmobile_ohci *)hcd_to_ohci(hcd);
	ohci->phys_base = r->start;
	ohci->base = hcd->regs;
	ohci->pdata = pdata;
	if (pdata->oci_irq) {
		if (!pdata->check_oci) {
			retval = -ENODEV;
			goto err3;
		} else {
			retval = request_irq(pdata->oci_irq,
				rmobile_over_current_irq,
				IRQF_TRIGGER_RISING | IRQF_TRIGGER_FALLING |
					IRQF_SHARED, "Over Current", hcd);
			if (!retval)
				goto err3;
			disable_irq(pdata->oci_irq);
		}
	}

	pm_suspend_ignore_children(&pdev->dev, true);
	pm_runtime_enable(&pdev->dev);
	pm_runtime_get_noresume(&pdev->dev);

	rmobile_hc_start(hcd);
	ohci_hcd_init(hcd_to_ohci(hcd));

	retval = usb_add_hcd(hcd, irq, IRQF_SHARED);
	if (retval == 0)
		return retval;

	rmobile_hc_stop(ohci);

	pm_runtime_disable(&pdev->dev);

	if (pdata->oci_irq)
		free_irq(pdata->oci_irq, hcd);
err3:
	iounmap(hcd->regs);
err2:
	release_mem_region(hcd->rsrc_start, hcd->rsrc_len);
err1:
	usb_put_hcd(hcd);

	return retval;
}

static void rmobile_ohci_remove(struct usb_hcd *hcd,
			struct platform_device *pdev)
{
	struct rmobile_ohci *ohci = to_rmobile_ohci(hcd);

	usb_remove_hcd(hcd);
	rmobile_hc_stop(ohci);
	if (ohci->pdata->oci_irq)
		free_irq(ohci->pdata->oci_irq, hcd);
	iounmap(ohci->base);
	release_mem_region(hcd->rsrc_start, hcd->rsrc_len);
	usb_put_hcd(hcd);
	pm_runtime_disable(&pdev->dev);
}

/*-------------------------------------------------------------------------*/

static int __devinit ohci_rmobile_start(struct usb_hcd *hcd)
{
	struct ohci_hcd *ohci = hcd_to_ohci(hcd);
	int ret;

	if ((ret = ohci_init(ohci)) < 0)
		return ret;

	if ((ret = ohci_run(ohci)) < 0) {
		err("can't start %s", hcd->self.bus_name);
		ohci_stop(hcd);
		return ret;
	}

	return 0;
}

/*-------------------------------------------------------------------------*/

static struct hc_driver ohci_rmobile_hc_driver = {
	.description		= hcd_name,
	.product_desc		= "R-Mobile OHCI",
	.hcd_priv_size		= sizeof(struct rmobile_ohci),

	/*
	 * generic hardware linkage
	 */
	.irq			= ohci_irq,
	.flags			= HCD_USB11 | HCD_MEMORY,

	/*
	 * basic lifecycle operations
	 */
	.start			= ohci_rmobile_start,
	.stop			= ohci_stop,
	.shutdown		= ohci_shutdown,

	/*
	 * managing i/o requests and associated device resources
	 */
	.urb_enqueue		= ohci_urb_enqueue,
	.urb_dequeue		= ohci_urb_dequeue,
	.endpoint_disable	= ohci_endpoint_disable,

	/*
	 * scheduling support
	 */
	.get_frame_number	= ohci_get_frame,

	/*
	 * root hub support
	 */
	.hub_status_data	= ohci_hub_status_data,
	.hub_control		= rmobile_hub_control,
#ifdef CONFIG_PM
	.bus_suspend		= ohci_bus_suspend,
	.bus_resume		= ohci_bus_resume,
#endif
	.start_port_reset	= ohci_start_port_reset,
};

/*-------------------------------------------------------------------------*/

static int ohci_hcd_drv_probe(struct platform_device *pdev)
{
	if (usb_disabled())
		return -ENODEV;

	return rmobile_ohci_probe(&ohci_rmobile_hc_driver, pdev);
}

static int ohci_hcd_drv_remove(struct platform_device *pdev)
{
	struct usb_hcd *hcd = platform_get_drvdata(pdev);

	rmobile_ohci_remove(hcd, pdev);
	return 0;
}

#ifdef CONFIG_PM
static int ohci_hcd_drv_suspend(struct device *dev)
{
	struct usb_hcd *hcd = dev_get_drvdata(dev);
	struct rmobile_ohci *ohci = to_rmobile_ohci(hcd);

	if (ohci->pdata->oci_irq)
		disable_irq(ohci->pdata->oci_irq);

	rmobile_vbus_control(ohci, 0);
	rmobile_ohci_save_register(hcd);
	rmobile_hc_stop(ohci);

	pm_runtime_put(dev);

	return 0;
}

static int ohci_hcd_drv_resume(struct device *dev)
{
	struct usb_hcd *hcd = dev_get_drvdata(dev);
	struct rmobile_ohci *ohci = to_rmobile_ohci(hcd);

	pm_runtime_get_sync(dev);

	rmobile_hc_start(hcd);
	rmobile_ohci_load_register(hcd);
	rmobile_vbus_control(ohci, 1);

	ohci_finish_controller_resume(hcd);
	return 0;
}

static const struct dev_pm_ops ohci_hcd_pm_ops = {
	.suspend	= ohci_hcd_drv_suspend,
	.resume		= ohci_hcd_drv_resume,
};
#endif

static struct platform_driver ohci_hcd_rmobile_driver = {
	.probe		= ohci_hcd_drv_probe,
	.remove		= ohci_hcd_drv_remove,
	.shutdown	= usb_hcd_platform_shutdown,
	.driver		= {
		.name	=  "rmobile-ohci-driver",
		.owner	= THIS_MODULE,
#ifdef CONFIG_PM
		.pm	= &ohci_hcd_pm_ops,
#endif
	},
};

MODULE_ALIAS("platform:rmobile-ohci-driver");
