#include <linux/kernel.h>
#include <linux/platform_device.h>
#include <linux/interrupt.h>
#include <linux/irq.h>
#include <linux/io.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/fb.h>
#include <linux/clk.h>

#include <mach/irqs.h>
#include <mach/rmobile_hdmi.h>

#ifdef DEBUG
#define d_dbg(fmt, args...) { if (1) printk(fmt, ##args); 0; }
#else
#define d_dbg(fmt, args...) { if (0) printk(fmt, ##args); 0; }
#endif

#ifdef TRACE_DEBUG
#define trace() { if (1) printk("*** %s()\n", __func__); 0; }
#else
#define trace() { if (0) printk("*** %s()\n", __func__); 0; }
#endif

struct sh_mobile_hdmi_priv {
	u32 *base;
	int irq;

	struct notifier_block nb;
	int blanked;
};

struct hdmi_vid {
	u32 vid;
	u32 rate;
	u32 xres;
	u32 yres;
	u32 refresh;
	u32 vmode;
};

static struct sh_mobile_hdmi_priv *hdmi_priv;
static struct hdmi_vid vid_modes[] = {
	/*  480p@59.94Hz */  [2] = {  2,  27000000,  720,  480, 59, 0 },
	/*  720p@60Hz */  [4] = {  4,  74250000, 1280,  720, 60, 0 },
	/* 1080i@60Hz */  [5] = {  5,  74250000, 1920, 1080, 60, FB_VMODE_INTERLACED },
	/* 1080p@60Hz */ [16] = { 16, 148500000, 1920, 1080, 60, 0 },
	/* 1080p@24Hz */ [32] = { 32,  74250000, 1920, 1080, 24, 0 },
};
static u32 current_vid = 5;
static u32 current_hdmi = 0;
static const char *source_dev_name;

static struct work_struct hdmi_work;

static DECLARE_MUTEX(hdmi_sem);

static int update_clock(u32 rate)
{
	struct clk *hdmi_clk = clk_get(NULL, "hdmi_clk");
	struct clk *parent;
	char *parent_name;

	switch (rate) {
	case 27000000:
		parent_name = "dv_clk";
		break;
	case 148500000:
		parent_name = "pllc1_div2_clk";
		break;
	case 74250000:
	default:
		parent_name = "pllc1_div2_clk";
		rate = 74250000;
		break;
	}

	clk_disable(hdmi_clk);
	parent = clk_get(NULL, parent_name);
	clk_set_parent(hdmi_clk, parent);
	clk_set_rate(hdmi_clk, rate);
	clk_enable(hdmi_clk);
	clk_put(hdmi_clk);
	clk_put(parent);

	return 0;
}

static u32 find_vid_mode(struct fb_videomode *mode)
{
	int i;

	for (i=0; i<ARRAY_SIZE(vid_modes); i++) {
		if (vid_modes[i].xres == mode->xres &&
		    vid_modes[i].yres == mode->yres &&
		    vid_modes[i].refresh == mode->refresh &&
		    vid_modes[i].vmode == mode->vmode)
			return vid_modes[i].vid;
	}

	return -1;
}

#define hdmi_delay(x) mdelay(x)

static u32 hdmi_read(u32 offset)
{
	struct sh_mobile_hdmi_priv *priv = hdmi_priv;
	u32 val;

	val = ioread32(priv->base + offset);
	hdmi_delay(1);

	return val;
}

static void hdmi_write(u32 offset, u32 val)
{
	struct sh_mobile_hdmi_priv *priv = hdmi_priv;

	iowrite32(val, priv->base + offset);
	hdmi_delay(1);
}
#define hdmi_phy_write(offset, val) hdmi_write(offset, val)

static void setup_hdmi(u32 vid, int interlace, int hdmi)
{
	struct sh_mobile_hdmi_priv *priv = hdmi_priv;
	u32 rate = vid_modes[vid].rate;
	u32 val;

	trace();

	down(&hdmi_sem);
	disable_irq(priv->irq);

	update_clock(74250000);

	/* HDMI with Audio */
	hdmi_write(0x00, 0x23);
	if (hdmi)
		hdmi_write(0xaf, 0x12);
	else
		hdmi_write(0xaf, 0x10);
	hdmi_write(0x88, 0xff);
	hdmi_write(0x89, 0xff);
	hdmi_phy_write(0x103f, 0x00000000);
	hdmi_phy_write(0x1004, 0x0000000b);
	hdmi_phy_write(0x1048, 0x00006710);
	hdmi_phy_write(0x1051, 0x01020406);
	hdmi_phy_write(0x1052, 0x07080806);
	hdmi_phy_write(0x1053, 0x01020406);
	hdmi_phy_write(0x1054, 0x07080806);
	hdmi_phy_write(0x1055, 0x01020406);
	hdmi_phy_write(0x1056, 0x07080806);
	hdmi_phy_write(0x1057, 0x01020406);
	hdmi_phy_write(0x1058, 0x07080806);
	hdmi_phy_write(0x1059, 0x01020406);
	hdmi_phy_write(0x105a, 0x07080806);
	hdmi_phy_write(0x105b, 0x01020406);
	hdmi_phy_write(0x105c, 0x07080806);
	hdmi_phy_write(0x105d, 0x00000000);
	hdmi_phy_write(0x105e, 0x00000000);
	hdmi_phy_write(0x105f, 0x00000000);
	hdmi_phy_write(0x1060, 0x00000000);
	hdmi_phy_write(0x1061, 0x00000000);
	hdmi_phy_write(0x1062, 0x00000000);
	hdmi_write(0x5f, 0x06);
	hdmi_write(0x00, 0x43);
	if (current_vid == -1)
		return;

	if (interlace)
		hdmi_write(0x30, 0x62);
	else
		hdmi_write(0x30, 0x60);
	hdmi_write(0x5f, 0x06);
	hdmi_write(0x60, 0x82);
	hdmi_write(0x61, 0x02);
	hdmi_write(0x62, 0x0d);
	hdmi_write(0x64, 0x00);
	hdmi_write(0x65, 0x08);
	hdmi_write(0x66, 0x00);
	hdmi_write(0x67, vid);
	hdmi_write(0x68, 0x00);
	hdmi_write(0x69, 0x00);
	hdmi_write(0x6a, 0x00);
	hdmi_write(0x6b, 0x00);
	hdmi_write(0x6c, 0x00);
	hdmi_write(0x6d, 0x00);
	hdmi_write(0x6e, 0x00);
	hdmi_write(0x6f, 0x00);
	hdmi_write(0x70, 0x00);
	hdmi_write(0x71, 0x00);
	hdmi_write(0x72, 0x00);
	hdmi_write(0x73, 0x00);
	hdmi_write(0x74, 0x00);
	hdmi_write(0x75, 0x00);
	hdmi_write(0x76, 0x00);
	hdmi_write(0x77, 0x00);
	hdmi_write(0x78, 0x00);
	hdmi_write(0x63, 0x63);
	hdmi_write(0x01, 0x00);
	hdmi_write(0x02, 0x18);
	hdmi_write(0x03, 0x00);
	hdmi_write(0x04, 0x20);
	hdmi_write(0x0a, 0x08);
	hdmi_write(0x0b, 0x40);
	hdmi_write(0x0c, 0x04);
	hdmi_write(0x0d, 0x00);
	hdmi_write(0x10, 0x00);
	hdmi_write(0x11, 0x00);
	hdmi_write(0x12, 0x00);
	hdmi_write(0x13, 0x00);
	hdmi_write(0x14, 0x00);
	hdmi_write(0x15, 0x20);
	hdmi_write(0xd6, 0x18);
	hdmi_write(0xd5, 0x00);
	hdmi_write(0x5f, 0x08);
	hdmi_write(0x60, 0x84);
	hdmi_write(0x61, 0x01);
	hdmi_write(0x62, 0x0a);
	hdmi_write(0x64, 0x00);
	hdmi_write(0x65, 0x00);
	hdmi_write(0x66, 0x00);
	hdmi_write(0x67, 0x00);
	hdmi_write(0x68, 0x00);
	hdmi_write(0x69, 0x00);
	hdmi_write(0x6a, 0x00);
	hdmi_write(0x6b, 0x00);
	hdmi_write(0x6c, 0x00);
	hdmi_write(0x6d, 0x00);
	hdmi_write(0xd3, 0x01);
	hdmi_write(0x3b, 0x00);
	hdmi_write(0x45, 0x40);
	hdmi_write(0x15, 0x20);
	hdmi_write(0x92, 0x9f);
	hdmi_write(0x16, 0x04);
	hdmi_write(0x17, 0x22);
	hdmi_phy_write(0x1014, 0x00000008);
	hdmi_phy_write(0x1009, 0x00000000);
	hdmi_phy_write(0x100a, 0x00000000);
	hdmi_phy_write(0x1016, 0x00000000);
	hdmi_phy_write(0x100b, 0x00000081);
	hdmi_phy_write(0x1006, 0x00000000);
	hdmi_phy_write(0x100d, 0x0000000f);
	hdmi_phy_write(0x100e, 0x0000000f);
	hdmi_phy_write(0x1010, 0x00000003);
	hdmi_phy_write(0x1011, 0x00000001);
	if (rate == 27000000)
		hdmi_phy_write(0x100f, 0x00000012);
	else
		hdmi_phy_write(0x100f, 0x00000016);
	hdmi_delay(1);
	hdmi_phy_write(0x1010, 0x00000001);
	hdmi_delay(1);
	hdmi_phy_write(0x1010, 0x00000003);
	hdmi_phy_write(0x1011, 0x00000001);
	hdmi_phy_write(0x100d, 0x0000000f);
	hdmi_phy_write(0x100e, 0x0000000f);
	hdmi_write(0x00, 0x83);
	hdmi_write(0x40, 0x80);
	hdmi_write(0x41, 0x00);
	hdmi_write(0x45, 0x80);
	hdmi_write(0x40, 0x80);
	hdmi_write(0x41, 0x00);
	hdmi_write(0xfc, 0x00);
	hdmi_write(0xf3, 0x00);
	hdmi_write(0xf4, 0x00);
	hdmi_write(0xf5, 0x00);
	hdmi_write(0xf6, 0x00);
	hdmi_write(0xf7, 0x00);
	hdmi_write(0xf8, 0x00);
	hdmi_write(0xf9, 0x00);
	hdmi_write(0xfa, 0x00);
	hdmi_write(0xfb, 0x00);
	hdmi_phy_write(0x1002, 0x00000000);
	hdmi_phy_write(0x1007, 0x00000000);
	hdmi_phy_write(0x1008, 0x00000000);
	hdmi_phy_write(0x1020, 0x0000000a);

	val = hdmi_read(0xdf);
	if (val & 0x80)
		hdmi_write(0x00, 0x83);
	else
		hdmi_write(0x00, 0x43);

	update_clock(rate);

	enable_irq(priv->irq);
	up(&hdmi_sem);
}

static void dump_edid(unsigned char *edid)
{
#ifdef DEBUG
	int i;

	for (i=0; i<128; i++) {
		printk("%02x ", edid[i]);
		if (i%16 == 15)
			printk("\n");
	}
	if (i%16 != 0)
		printk("\n");
#endif
}

static void sh_mobile_hdmi_read_edid(void)
{
	u32 stat1, stat2;
	int i, j;
	u8 edid_1st[256];
	u8 edid[256];
	int retry = 10;

	trace();

	update_clock(74250000);

	current_hdmi = 0;
retry1:
	hdmi_write(0xc5, 0x00);
	hdmi_write(0xc4, 0x00);

	/* Waiting for EDID Read Ready. */
	for (i=0; i<100; i++) {
		stat1 = hdmi_read(0x94);
		if (stat1 & 0x06)
			break;
	}

	if (stat1 & 0x02) {
		if (--retry > 0)
			goto retry1;
		d_dbg("failed to read EDID.\n");
		return;
	}

	if (stat1 & 0x04) {
		for (i=0; i<128; i++) {
			edid_1st[i] = hdmi_read(0x80);
			hdmi_read(0x80);	/* dummy read */
		}
		d_dbg("EDID[0]:\n");
		dump_edid(edid_1st);

		for (j=1; j<edid_1st[126] + 1; j++) {
			u32 seg;
			u32 blk;

			seg = j / 2;
			blk = (j % 2) * 0x80;
retry2:
			hdmi_write(0xc5, blk);
			hdmi_write(0xc4, seg);

			do {
				stat1 = hdmi_read(0x94);
				stat2 = hdmi_read(0x95);
			} while (!(stat1 & 0x06));

			if (stat1 & 0x02)
				goto retry2;
			if (stat1 & 0x04) {
				for (i=0; i<128; i++) {
					edid[i] = hdmi_read(0x80);
					hdmi_read(0x80);	/* dummy read */
				}
				if (edid[1] == 3)
					current_hdmi = 1;

				d_dbg("EDID[%d]:\n", j);
				dump_edid(edid);
			}
		}
	}
	if (current_vid != -1)
		update_clock(vid_modes[current_vid].rate);
}

static void sh_mobile_hdmi_setup(struct work_struct *work)
{
	struct sh_mobile_hdmi_priv *priv = hdmi_priv;

	trace();

	if (priv->blanked)
		return;

	if (current_vid != -1)
		setup_hdmi(current_vid,
			   vid_modes[current_vid].vmode & FB_VMODE_INTERLACED,
			   current_hdmi);
}

static irqreturn_t sh_mobile_hdmi_irq(int irq, void *dev_id)
{
	u32 stat1, stat2;
	u32 val;

	hdmi_write(0x00, 0x23);

	stat1 = hdmi_read(0x94);
	stat2 = hdmi_read(0x95);

	if (!stat1 && !stat2) {
		hdmi_write(0x00, 0x23);
		return IRQ_HANDLED;
	}

	/* Clear interrupt status */
	hdmi_write(0x94, 0xff);
	hdmi_write(0x95, 0xff);

	if (stat1 & 0x80) {
		mdelay(10);
		val = hdmi_read(0xdf);
		/* printk("0xe6be037c = %08x\n", val); */
		if (val & 0x80) {
			sh_mobile_hdmi_read_edid();
			schedule_work(&hdmi_work);
		} else {
			hdmi_write(0x00, 0x23);
		}
	}

	return IRQ_HANDLED;
}

static int sh_mobile_hdmi_event_handler(struct notifier_block *nb,
					unsigned long action, void *data)
{
	struct sh_mobile_hdmi_priv *priv = hdmi_priv;
	struct fb_event *event = data;
	struct fb_info *info = event->info;
	struct fb_videomode *mode;

	trace();

	if (strncmp(source_dev_name, dev_name(info->device),
		    strlen(source_dev_name)))
		return 0;

	switch (action) {
	case FB_EVENT_MODE_CHANGE:
		mode = event->data;
		current_vid = find_vid_mode(mode);
		if (current_vid != -1)
			schedule_work(&hdmi_work);
		else
			hdmi_write(0x00, 0x23);
		break;
	case FB_EVENT_BLANK:
		if (*(int *)event->data)
			priv->blanked = 1; /* blank */
		else
			priv->blanked = 0; /* unblank */
		break;
	default:
		break;
	}
	return 0;
}

static int __devinit sh_mobile_hdmi_probe(struct platform_device *pdev)
{
	struct sh_mobile_hdmi_info *pdata = pdev->dev.platform_data;
	struct sh_mobile_hdmi_priv *priv;
	struct resource *res;
	int ret;
	int irq;

	trace();

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	irq = platform_get_irq(pdev, 0);
	if (!res || irq < 0) {
		dev_err(&pdev->dev, "cannot get platform resources\n");
		return -ENOENT;
	}

	hdmi_priv = kzalloc(sizeof(struct sh_mobile_hdmi_priv), GFP_KERNEL);
	if (!hdmi_priv) {
		dev_err(&pdev->dev, "unable to allocate memory\n");
		return -ENOMEM;
	}

	priv = hdmi_priv;
	priv->base = ioremap_nocache(res->start, (res->end - res->start) + 1);
	priv->irq = irq;

	/* PLL reset */
	hdmi_write(0x00, 0x2c);
	hdmi_write(0x00, 0x23);

	source_dev_name = pdata->source_dev_name;
	INIT_WORK(&hdmi_work, sh_mobile_hdmi_setup);

	/* PHY setup: "vid = 5" is dummy value */
	setup_hdmi(5, 0, 0);

	ret = request_irq(irq, sh_mobile_hdmi_irq, 0, dev_name(&pdev->dev),
			  NULL);
	if (ret) {
		iounmap(priv->base);
		kfree(priv);
		return -EBUSY;
	}

	priv->nb.notifier_call = sh_mobile_hdmi_event_handler;
	fb_register_client(&priv->nb);

	return 0;
}

static int sh_mobile_hdmi_remove(struct platform_device *pdev)
{
	struct sh_mobile_hdmi_priv *priv = hdmi_priv;

	trace();

	fb_unregister_client(&priv->nb);
	free_irq(priv->irq, NULL);
	cancel_work_sync(&hdmi_work);
	iounmap(priv->base);
	kfree(hdmi_priv);

	return 0;
}

static struct platform_driver sh_mobile_hdmi_driver = {
	.driver = {
		.name = "sh_mobile_hdmi",
		.owner = THIS_MODULE,
	},
	.probe = sh_mobile_hdmi_probe,
	.remove = sh_mobile_hdmi_remove,
};

static int __init rmobile_hdmi_init(void)
{
	trace();
	return platform_driver_register(&sh_mobile_hdmi_driver);
}

static void __exit rmobile_hdmi_exit(void)
{
	trace();
	platform_driver_unregister(&sh_mobile_hdmi_driver);
}

module_init(rmobile_hdmi_init);
module_exit(rmobile_hdmi_exit);

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("R-Mobile A1 HDMI driver");
