/*
 *  File Name       : RSPF0_API.c
 *  Function        : SPU OPEN SOUND I/F
 *  Release Version : Ver 1.00
 *  Release Date    : 2011/11/30
 *
 *  Copyright (C) 2011 Renesas Electronics Corporation
 *
 *  This program is free software;you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by Free
 *  Softwere Foundation; either version 2 of License, or (at your option) any
 *  later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY;
 *  without even the implied warrnty of MERCHANTABILITY or FITNESS FOR A
 *  PARTICULAR PURPOSE.
 *  See the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program;
 *  If not, write to the Free Software Foundation, Inc., 51 Franklin St,
 *  Fifth Floor, Boston, MA  02110-1301  USA
 *
 */
#include "../spu_general_api/RSPF0_spu_api.h"
#include "../spu_general_api/RSPF0_spu_api_system.h"
#include "../spu_general_api/RSPF0_spu_api_dataport.h"
#include "../spu_general_api/RSPF0_spu_api_fmsiport.h"
#include "../spu_general_api/RSPF0_spu_api_soundv2.h"




#include "../include/RSPF0_API_S0.h"

#include "RSPF0.h"

/* definition */
#define RSPF0_VERSION	0x30000216

#define RSPF0_DSP_MAX     2
#define RSPF0_PATH_NUM    3
#define RSPF0_API_TIMEOUT 1000
#define RSPF0_BUFFER_SIZE 128

enum {
	RSPF0_STATE_INIT = 0,
	RSPF0_STATE_OPEN
};

enum {
	RSPF0_PORT_STATE_INIT = 0,
	RSPF0_PORT_STATE_DISABLE,
	RSPF0_PORT_STATE_ENABLE
};

enum {
	RSPF0_SOURCE = 0,
	RSPF0_DEST   = 1
};

/* local variables */
static RSPF0_Path *rspf0_path[RSPF0_DSP_MAX][RSPF0_PATH_NUM];

static void RSPF0_SourceTransferEndCallBack(sSPUDataPort *port,
				unsigned long size, unsigned long flush);
static void RSPF0_DestTransferEndCallBack(sSPUDataPort *port,
				unsigned long size, unsigned long flush);
static void RSPF0_FmsiPortErrorCallBack(sSPUFmsiPort *port,
				unsigned long info, unsigned long flush);

static long rspf0_open_data_port(RSPF0_Path *path, RSPF0_OpenParam *param,
						unsigned long port_type);
static long rspf0_close_data_port(RSPF0_Path *path, unsigned long port_type);
static long rspf0_open_fmsi_port(RSPF0_Path *path, RSPF0_OpenParam *param,
						unsigned long port_type);
static long rspf0_close_fmsi_port(RSPF0_Path *path, unsigned long port_type);
static long rspf0_open_sound(RSPF0_Path *path, RSPF0_OpenParam *param);
static long rspf0_close_sound(RSPF0_Path *path);

static long get_path_index(RSPF0_Path *path);

/* FMSI_IP_ONLY Start */
unsigned long rspf0_fmsiip;			/* data for FMSI IP */
#define RST1IF_FMSI_ADDRESS (0x00000100)	/* offset from DSP IO */
#define RST1IF_FMSI_VECTOR  (0x00000001)	/* ISR2 */
/* bit of INT_ST/IEMSK/IMSK for FMSI  */
enum {
	RST1IF_FMSI_SOUT = 0,
	RST1IF_FMSI_SIN  = 4,
	RST1IF_FMSI_POUT = 12,		/* Dummy(Reaserv bit) */
	RST1IF_FMSI_PIN  = 8
};
/* FMSI_IP_ONLY END */


/*-------------------------------------------------------
Name    : RSPF0_Init()
Outline : The download of this middleware and the start
		processing of this middleware are done.
Input   : dsp_id        : used dsp number,
	: *func         : ptr of RSPF0_Func,
	: address_mode  : Forwarding address of DMA
Output  : return code
-------------------------------------------------------*/
long RSPF0_Init(unsigned long dsp_id,
		RSPF0_Func *func,
		unsigned long address_mode) {
	const unsigned long *data[3] = {
					(const unsigned long *) RSPF0_PMEM_DATA,
					(const unsigned long *) RSPF0_XMEM_DATA,
					(const unsigned long *) RSPF0_YMEM_DATA
	};
	const unsigned long size[3] = {	RSPF0_PMEM_SIZE,
					RSPF0_XMEM_SIZE,
					RSPF0_YMEM_SIZE
	};
	unsigned long i;
	unsigned long capacity[3];
	SPUAPI_Func spuapi_func;
	long result;

	/* parameter check */
	if (dsp_id >= RSPF0_DSP_MAX || !func)
		/* check other parameters in SPUAPI_Open()*/
		return RSPF0_RTN_ERROR_PARAM;

	spuapi_func.reg_write = func->reg_write;
	spuapi_func.reg_read = func->reg_read;
	spuapi_func.set_event = func->set_event;
	spuapi_func.wait_for_event = func->wait_for_event;
	spuapi_func.enter_critical_section = func->enter_critical_section;
	spuapi_func.leave_critical_section = func->leave_critical_section;
	capacity[0] = RSPF0_PBANK_NUM * 8192;
	capacity[1] = RSPF0_XBANK_NUM * 8192;
	capacity[2] = RSPF0_YMEMORY_SIZE;

	result = SPUAPI_Init(dsp_id);	/* It ends normally without fail. */
	result = SPUAPI_Open(dsp_id, &spuapi_func, data, size, capacity,
		address_mode);
	if (result != RSPF0_RTN_SUCCESS) {
		if (result == -1)
			return RSPF0_RTN_ERROR_PARAM;
		else
			return RSPF0_RTN_ERROR_DRV;
	}
	if (SPUSystem_Init(dsp_id) != RSPF0_RTN_SUCCESS)
		return RSPF0_RTN_ERROR_DRV;
	if (SPUDataPort_Init(dsp_id) != RSPF0_RTN_SUCCESS)
		return RSPF0_RTN_ERROR_DRV;
	if (SPUFmsiPort_Init(dsp_id) != RSPF0_RTN_SUCCESS)
		return RSPF0_RTN_ERROR_DRV;
	if (SPUSound_Init(dsp_id) != RSPF0_RTN_SUCCESS)
		return RSPF0_RTN_ERROR_DRV;
	for (i = 0; i < RSPF0_PATH_NUM; i++)
		rspf0_path[dsp_id][i] = 0;

	/* FMSI_IP_ONLY Start */
	result = SPUFmsiPort_Allocate(dsp_id, RST1IF_FMSI_ADDRESS,
					 RST1IF_FMSI_VECTOR, &rspf0_fmsiip,
					 SPUAPI_FMSIPORT_BITSELECT_FMSI);
	if (result != 0)
		return RSPF0_RTN_ERROR_DRV;
	/* FMSI_IP_ONLY END */

	return RSPF0_RTN_SUCCESS;
}


/*-------------------------------------------------------
Name    : RSPF0_Quit()
Outline : The termination of the middleware is done.
Input   : dsp_id        : used dsp number,
Output  : return code
-------------------------------------------------------*/
long RSPF0_Quit(unsigned long dsp_id)
{
	long result;

	rspf0_fmsiip = 0;	/* FMSI_IP_ONLY:Reset data for FMSI IP */

	result = SPUDRV_FUNC(Close)(dsp_id);
		/* check max value of dsi_id in SPUDRV_FUNC(Close)(dsp_id) */
	if (result == RSPF0_RTN_SUCCESS)
		return RSPF0_RTN_SUCCESS;
	else if (result == -1)
		return RSPF0_RTN_ERROR_PARAM;
	else if (result == -2)
		return RSPF0_RTN_ERROR_STATE;
	else
		return RSPF0_RTN_ERROR_DRV;
}

/*-------------------------------------------------------
Name    : RSPF0_Interrupt()
Outline : The interrupt processing of the middleware
		is executed.
Input   : dsp_id        : used dsp number
Output  : None
-------------------------------------------------------*/
void RSPF0_Interrupt(unsigned long dsp_id)
{
	SPUAPI_Interrupt(dsp_id);
}


/*-------------------------------------------------------
Name    : RSPF0_Open()
Outline : creating and the initialization of the work
		area that this middleware uses are done.
Input   : dsp_id      : used dsp number,
	: *path       : ptr of RSPF0_Path
			(keep setting param),
	: *param      : ptr of RSPF0_OpenParam
			(set param)
Output  : return code
-------------------------------------------------------*/
long RSPF0_Open(unsigned long dsp_id, RSPF0_Path *path,
						RSPF0_OpenParam *param) {
	int i, index;
	unsigned long in_ch;
	long result;
	long result_del;

	/* Parameter check */
	if (dsp_id >= RSPF0_DSP_MAX || path == 0 || param == 0)
		return RSPF0_RTN_ERROR_PARAM;
	if (param->swap_mode > RSPF0_SWAP_MODE_BYTE_WORD ||
	param->source_port >= RSPF0_PORT_MAX ||
	param->dest_port >= RSPF0_PORT_MAX ||
	(param->source_port != RSPF0_PORT_CPU &&
	param->dest_port != RSPF0_PORT_CPU) ||
	(!(param->source_port == RSPF0_PORT_CPU &&
	param->dest_port == RSPF0_PORT_CPU) &&
	param->fmsi_error_callback == 0) ||
	(param->source_port == RSPF0_PORT_CPU &&
	param->source_end_callback == 0) ||
	(param->dest_port == RSPF0_PORT_CPU && param->dest_end_callback == 0) ||
	(param->channel == 0x00 || param->channel > RSPF0_CHANNEL_STEREO)
	)
		return RSPF0_RTN_ERROR_PARAM;

	/* resource check */
	for (i = 0; i < RSPF0_PATH_NUM; i++) {
		if (rspf0_path[dsp_id][i] == 0)
			break;					/* pgr0701 */
	}
	if (i == RSPF0_PATH_NUM)
		return RSPF0_RTN_ERROR_RESOURCE;
	index = i;

	path->dsp_id = dsp_id;

	/* Init Port Status */
	for (i = 0; i < 2; i++)
		path->fmsi_port_status[i] = RSPF0_PORT_STATE_INIT;

	/* Buffer */
	result = SPUSystem_CreateBuffer(dsp_id, RSPF0_BUFFER_SIZE,
		SPUAPI_BUFFER_TYPE_WORD_RING, &path->buffer[RSPF0_SOURCE]);
	if (result == RSPF0_RTN_ERROR_PARAM)
		return RSPF0_RTN_ERROR_PARAM;
	if (result != RSPF0_RTN_SUCCESS)
		return RSPF0_RTN_ERROR_RESOURCE;
	result = SPUSystem_CreateBuffer(dsp_id, RSPF0_BUFFER_SIZE,
		SPUAPI_BUFFER_TYPE_WORD_RING, &path->buffer[RSPF0_DEST]);
	if (result != RSPF0_RTN_SUCCESS) {
		result_del = SPUSystem_DeleteBuffer
						(&path->buffer[RSPF0_SOURCE]);
		if (result_del != RSPF0_RTN_SUCCESS)
			return RSPF0_RTN_ERROR_FATAL;	/* misstake to delete */
		if (result == RSPF0_RTN_ERROR_PARAM)
			return RSPF0_RTN_ERROR_PARAM;
		else
			return RSPF0_RTN_ERROR_RESOURCE;
	}
	/* Source(Input) */
	if (param->source_port == RSPF0_PORT_CPU) {
		result = rspf0_open_data_port(path, param, RSPF0_SOURCE);
		if (result != RSPF0_RTN_SUCCESS)
			goto RSPF0_Open_Error1;
	} else{
		result = rspf0_open_fmsi_port(path, param, RSPF0_SOURCE);
		if (result != RSPF0_RTN_SUCCESS)
			goto RSPF0_Open_Error1;
	}
	/* Destination(Output) */
	if (param->dest_port == RSPF0_PORT_CPU) {
		result = rspf0_open_data_port(path, param, RSPF0_DEST);
		if (result != RSPF0_RTN_SUCCESS)
			goto RSPF0_Open_Error2;
	} else{
		in_ch = param->channel;
		result = rspf0_open_fmsi_port(path, param, RSPF0_DEST);
		if (result != RSPF0_RTN_SUCCESS)
			goto RSPF0_Open_Error2;
		param->channel = in_ch;		/* reteunr input channel */
	}
	/* Sound */
	result = rspf0_open_sound(path, param);
	if (result != RSPF0_RTN_SUCCESS)
		goto RSPF0_Open_Error3;
	/* Save Setting parameters */
	rspf0_path[dsp_id][index] = path;
	path->status               = RSPF0_STATE_OPEN;
	path->source_port          = param->source_port;
	path->dest_port            = param->dest_port;
	path->source_end_callback  = param->source_end_callback;
	path->dest_end_callback    = param->dest_end_callback;
	path->fmsi_error_callback  = param->fmsi_error_callback;

	return RSPF0_RTN_SUCCESS;

/*----------------------
	Error processing
----------------------*/
RSPF0_Open_Error3:
	if (param->dest_port == RSPF0_PORT_CPU)
		result_del = rspf0_close_data_port(path, RSPF0_DEST);
	else
		result_del = rspf0_close_fmsi_port(path, RSPF0_DEST);
	if (result_del != RSPF0_RTN_SUCCESS)
		return RSPF0_RTN_ERROR_FATAL;	/* misstake to delete */
RSPF0_Open_Error2:
	if (param->source_port == RSPF0_PORT_CPU)
		result_del = rspf0_close_data_port(path, RSPF0_SOURCE);
	else
		result_del = rspf0_close_fmsi_port(path, RSPF0_SOURCE);
	if (result_del != RSPF0_RTN_SUCCESS)
		return RSPF0_RTN_ERROR_FATAL;	/* misstake to delete */
RSPF0_Open_Error1:
	result_del = SPUSystem_DeleteBuffer(&path->buffer[RSPF0_DEST]);
	if (result_del != RSPF0_RTN_SUCCESS)
		return RSPF0_RTN_ERROR_FATAL;	/* misstake to delete */
	result_del = SPUSystem_DeleteBuffer(&path->buffer[RSPF0_SOURCE]);
	if (result_del != RSPF0_RTN_SUCCESS)
		return RSPF0_RTN_ERROR_FATAL;	/* misstake to delete */
	/* end deleting */
	return result;
}

/*-------------------------------------------------------
Name    : RSPF0_Close()
Outline : This middleware is stopped, and the work area
		used is released.
Input   : *path       : ptr of RSPF0_Path
			(keep setting param)
Output  : return code
-------------------------------------------------------*/
long RSPF0_Close(RSPF0_Path *path)
{
	long index;
	long result;

	/* check error param */
	index = get_path_index(path);
	if (index < 0)
		return RSPF0_RTN_ERROR_PARAM;

	/* Delete */
	result = rspf0_close_sound(path);
	if (result != RSPF0_RTN_SUCCESS)
		return result;
	/* Source */
	if (path->source_port == RSPF0_PORT_CPU)
		result = rspf0_close_data_port(path, RSPF0_SOURCE);
	else
		result = rspf0_close_fmsi_port(path, RSPF0_SOURCE);
	if (result != RSPF0_RTN_SUCCESS)
		return result;
	/* Destination */
	if (path->dest_port == RSPF0_PORT_CPU)
		result = rspf0_close_data_port(path, RSPF0_DEST);
	else
		result = rspf0_close_fmsi_port(path, RSPF0_DEST);
	if (result != RSPF0_RTN_SUCCESS)
		return result;
	/* Delete BUF */
	result = SPUSystem_DeleteBuffer(&path->buffer[RSPF0_DEST]);
	if (result != RSPF0_RTN_SUCCESS)
		return RSPF0_RTN_ERROR_FATAL;
	result = SPUSystem_DeleteBuffer(&path->buffer[RSPF0_SOURCE]);
	if (result != RSPF0_RTN_SUCCESS)
		return RSPF0_RTN_ERROR_FATAL;

	/* reset structure */
	path->status = RSPF0_STATE_INIT;
	rspf0_path[path->dsp_id][index] = 0;

	return RSPF0_RTN_SUCCESS;
}


/*-------------------------------------------------------
Name    : RSPF0_FmsiStart()
Outline : The processing of FMSI begins.
Input   : *path       : ptr of RSPF0_Path
			(keep setting param)
	: sync        : Synchronization /
			Asynchronization
Output  : return code
-------------------------------------------------------*/
long RSPF0_FmsiStart(RSPF0_Path *path, unsigned long sync)
{
	long result;

	result = get_path_index(path);
	if (result < 0)
		return RSPF0_RTN_ERROR_PARAM;
	if (path->source_port == RSPF0_PORT_CPU &&
	path->dest_port == RSPF0_PORT_CPU)		/* check setting */
		return RSPF0_RTN_ERROR_SETTING;

	/* Beginning of input port */
	if (path->source_port != RSPF0_PORT_CPU) {
		if (path->fmsi_port_status[RSPF0_SOURCE] ==
		RSPF0_PORT_STATE_DISABLE) {
			/* RESET FMSI Port */
			result = SPUFmsiPort_ResetError(&path->fmsi_port
				[RSPF0_SOURCE], sync);
			if (result == RSPF0_RTN_ERROR_PARAM)
				return RSPF0_RTN_ERROR_PARAM;
			if (result != RSPF0_RTN_SUCCESS)
				return RSPF0_RTN_ERROR_SPU_API;
			/* ENABLE FMSI Port */
			result = SPUFmsiPort_Enable(&path->fmsi_port
				[RSPF0_SOURCE], sync);
			if (result == RSPF0_RTN_ERROR_PARAM)
				return RSPF0_RTN_ERROR_PARAM;
			if (result != RSPF0_RTN_SUCCESS)
				return RSPF0_RTN_ERROR_SPU_API;
			path->fmsi_port_status[RSPF0_SOURCE] =
							RSPF0_PORT_STATE_ENABLE;
		} else{
			return RSPF0_RTN_ERROR_STATE;
		}
	}
	/* Beginning of output port */
	if (path->dest_port != RSPF0_PORT_CPU) {
		if (path->fmsi_port_status[RSPF0_DEST] ==
		RSPF0_PORT_STATE_DISABLE) {
			/* RESET FMSI Port */
			result = SPUFmsiPort_ResetError(&path->fmsi_port
				[RSPF0_DEST], sync);
			if (result == RSPF0_RTN_ERROR_PARAM)
				return RSPF0_RTN_ERROR_PARAM;
			if (result != RSPF0_RTN_SUCCESS)
				return RSPF0_RTN_ERROR_SPU_API;
			/* ENABLE FMSI Port */
			result = SPUFmsiPort_Enable(&path->fmsi_port
							[RSPF0_DEST], sync);
			if (result == RSPF0_RTN_ERROR_PARAM)
				return RSPF0_RTN_ERROR_PARAM;
			if (result != RSPF0_RTN_SUCCESS)
				return RSPF0_RTN_ERROR_SPU_API;
			path->fmsi_port_status[RSPF0_DEST] =
							RSPF0_PORT_STATE_ENABLE;
		} else{
			return RSPF0_RTN_ERROR_STATE;
		}
	}

	return RSPF0_RTN_SUCCESS;
}


/*-------------------------------------------------------
Name    : RSPF0_FmsiStop()
Outline : The processing of FMSI stops.
Input   : *path       : ptr of RSPF0_Path
			(keep setting param)
	: flush       : flushing flag of FMSI In
	: sync        : Synchronization /
			Asynchronization
Output  : return code
-------------------------------------------------------*/
long RSPF0_FmsiStop(RSPF0_Path *path, unsigned long flush, unsigned long sync)
{
	long result;

	result = get_path_index(path);
	if (result < 0)
		return RSPF0_RTN_ERROR_PARAM;
	if (flush > RSPF0_FLUSH_ON)
		return RSPF0_RTN_ERROR_PARAM;
	if (path->source_port == RSPF0_PORT_CPU &&
	path->dest_port == RSPF0_PORT_CPU)		/* check setting */
		return RSPF0_RTN_ERROR_SETTING;

	/* Stop of input port */
	if (path->source_port != RSPF0_PORT_CPU) {
		if (path->fmsi_port_status[RSPF0_SOURCE] ==
		RSPF0_PORT_STATE_ENABLE) {
			if (flush) {
				result = SPUFmsiPort_Flush(
					&path->fmsi_port[RSPF0_SOURCE], sync);
				if (result == RSPF0_RTN_ERROR_PARAM)
					return RSPF0_RTN_ERROR_PARAM;
				if (result != RSPF0_RTN_SUCCESS)
					return RSPF0_RTN_ERROR_SPU_API;
			} else{
				result = SPUFmsiPort_Disable(
					&path->fmsi_port[RSPF0_SOURCE], sync);
				if (result == RSPF0_RTN_ERROR_PARAM)
					return RSPF0_RTN_ERROR_PARAM;
				if (result != RSPF0_RTN_SUCCESS)
					return RSPF0_RTN_ERROR_SPU_API;
			}
			path->fmsi_port_status[RSPF0_SOURCE] =
						RSPF0_PORT_STATE_DISABLE;
		} else{
			return RSPF0_RTN_ERROR_STATE;
		}
	}
	/* Stop of output port */
	if (path->dest_port != RSPF0_PORT_CPU) {
		if (path->fmsi_port_status[RSPF0_DEST] ==
		RSPF0_PORT_STATE_ENABLE) {
			result = SPUFmsiPort_Disable(
					&path->fmsi_port[RSPF0_DEST], sync);
			if (result == RSPF0_RTN_ERROR_PARAM)
				return RSPF0_RTN_ERROR_PARAM;
			if (result != RSPF0_RTN_SUCCESS)
				return RSPF0_RTN_ERROR_SPU_API;
			path->fmsi_port_status[RSPF0_DEST] =
						RSPF0_PORT_STATE_DISABLE;
		} else{
			return RSPF0_RTN_ERROR_STATE;
		}
	}

	return RSPF0_RTN_SUCCESS;
}


/*-------------------------------------------------------
Name    : RSPF0_SourceTransferStart()
Outline : The data transfer on the input side of
		passing is begun.
Input   : *path       : ptr of RSPF0_Path
			(keep setting param)
	: source      : ptr of input data array
	: size        : transfer size in half_word
Output  : return code
-------------------------------------------------------*/
long RSPF0_SourceTransferStart(RSPF0_Path *path, short *source,
							unsigned long size) {
	long result;

	result = get_path_index(path);
	if (result < 0)
		return RSPF0_RTN_ERROR_PARAM;
	if (path->source_port != RSPF0_PORT_CPU)
		return RSPF0_RTN_ERROR_SETTING;
	if (!source)
		return RSPF0_RTN_ERROR_PARAM;

	result = SPUDataPort_TransferStart(&path->data_port[RSPF0_SOURCE],
				(unsigned long)source, size);	/* pgr0661 */
	if (result == RSPF0_RTN_ERROR_PARAM)
		return RSPF0_RTN_ERROR_PARAM;
	if (result != RSPF0_RTN_SUCCESS)
		return RSPF0_RTN_ERROR_SPU_API;

	return RSPF0_RTN_SUCCESS;
}


/*-------------------------------------------------------
Name    : RSPF0_DestTransferStart()
Outline : The data transfer on the output side of
		passing is begun.
Input   : *path       : ptr of RSPF0_Path
			(keep setting param)
	: source      : ptr of output data array
	: size        : transfer size in half_word
Output  : return code
-------------------------------------------------------*/
long RSPF0_DestTransferStart(RSPF0_Path *path, short *dest,
							unsigned long size) {
	long result;

	result = get_path_index(path);
	if (result < 0)
		return RSPF0_RTN_ERROR_PARAM;
	if (path->dest_port != RSPF0_PORT_CPU)
		return RSPF0_RTN_ERROR_SETTING;
	if (!dest)
		return RSPF0_RTN_ERROR_PARAM;
	if (size == 0)
		return RSPF0_RTN_ERROR_PARAM;

	result = SPUDataPort_TransferStart(&path->data_port[RSPF0_DEST],
				(unsigned long)dest, size);	/* pgr0661 */
	if (result == RSPF0_RTN_ERROR_PARAM)
		return RSPF0_RTN_ERROR_PARAM;
	if (result != RSPF0_RTN_SUCCESS)
		return RSPF0_RTN_ERROR_SPU_API;

	return RSPF0_RTN_SUCCESS;
}

/*-------------------------------------------------------
Name    : RSPF0_SetSrc()
Outline : Setting the parameter of SRC.
Input   : *path       : ptr of RSPF0_Path
			(keep setting param)
	: ratio       : ratio of SRC
	: sync        : Synchronization /
			Asynchronization
Output  : return code
-------------------------------------------------------*/
long RSPF0_SetSrc(RSPF0_Path *path, unsigned long ratio, unsigned long sync)
{
	long result;

	result = get_path_index(path);
	if (result < 0)
		return RSPF0_RTN_ERROR_PARAM;

	result = SPUSound_SetSRC(&path->sound, ratio, sync);
	if (result == RSPF0_RTN_ERROR_PARAM)
		return RSPF0_RTN_ERROR_PARAM;
	if (result != RSPF0_RTN_SUCCESS)
		return RSPF0_RTN_ERROR_SPU_API;

	return RSPF0_RTN_SUCCESS;
}

/*-------------------------------------------------------
Name    : RSPF0_GetVersion()
Outline : Refer this middleware version number.
Input   : none
Output  : version number
-------------------------------------------------------*/
unsigned long RSPF0_GetVersion(void)
{
	return RSPF0_VERSION;
}

/*-------------------------------------------------------
Name    : RSPF0_IoWrite()
Outline : Data is written in the FMSI register
		by way of SPU2.
Input   : dsp_id      : used dsp number,
	: addr        : Address of written FMSI register,
	: reg_data    : Written data
	: sync        : Synchronization /
			Asynchronization
Output  : return code
-------------------------------------------------------*/
long RSPF0_IoWrite(unsigned long dsp_id, unsigned long addr,
				unsigned long reg_value, unsigned long sync) {
	long result;

	if (sync > RSPF0_SPUAPI_WAIT_RESPONSE)
		/* check dsp_id, addr, reg_value in SPUSystem_IoWrite() */
		return RSPF0_RTN_ERROR_PARAM;
	result = SPUSystem_IoWrite(dsp_id, addr, reg_value, sync);
	if (result == RSPF0_RTN_ERROR_PARAM)
		return RSPF0_RTN_ERROR_PARAM;
	if (result != RSPF0_RTN_SUCCESS)
		return RSPF0_RTN_ERROR_SPU_API;
	return RSPF0_RTN_SUCCESS;
}


/*-------------------------------------------------------
Name    : RSPF0_IoRead()
Outline : Data is read in the FMSI register
		by way of SPU2.
Input   : dsp_id      : used dsp number,
	: addr        : Address of read FMSI register,
	: *reg_data   : ptr of Read data
	: sync        : Synchronization /
			Asynchronization
Output  : return code
-------------------------------------------------------*/
long RSPF0_IoRead(unsigned long dsp_id, unsigned long addr,
						unsigned long *reg_value) {
	long result;

	if (!reg_value)
		return RSPF0_RTN_ERROR_PARAM;
	result = SPUSystem_IoRead(dsp_id, addr, reg_value);
	if (result == RSPF0_RTN_ERROR_PARAM)
		return RSPF0_RTN_ERROR_PARAM;
	if (result != RSPF0_RTN_SUCCESS)
		return RSPF0_RTN_ERROR_SPU_API;
	return RSPF0_RTN_SUCCESS;
}


/*-------------------------------------------------------
Name    : RSPF0_Clear()
Outline : Clear internal buffers in SPU2.
Input   : *path       : ptr of RSPF0_Path
			(keep setting param)
	: sync        : Synchronization /
			Asynchronization
Output  : return code
Remarks : Not stop FMSIPort in this function.
		So user must call RSPF0_FMSIStop().
-------------------------------------------------------*/
long RSPF0_Clear(RSPF0_Path *path)
{
	long result;

	result = get_path_index(path);
	if (result < 0)
		return RSPF0_RTN_ERROR_PARAM;

	/* Stop except for FMSI */
	/* Guarantee all ptr in RSPF0_Open() and get_path_index() */
	result = SPUSound_Disable(&path->sound, RSPF0_SPUAPI_WAIT_RESPONSE);
	if (result != RSPF0_RTN_SUCCESS)
		return RSPF0_RTN_ERROR_SPU_API;
	if (path->source_port == RSPF0_PORT_CPU) {
		result = SPUDataPort_Disable(&path->data_port[RSPF0_SOURCE],
				0,	/* not read the rest of transfer */
				RSPF0_SPUAPI_WAIT_RESPONSE);
		if (result != RSPF0_RTN_SUCCESS)
			return RSPF0_RTN_ERROR_SPU_API;
	}
	if (path->dest_port == RSPF0_PORT_CPU) {
		result = SPUDataPort_Disable(&path->data_port[RSPF0_DEST],
				0,	/* not read the rest of transfer */
				RSPF0_SPUAPI_WAIT_RESPONSE);
		if (result != RSPF0_RTN_SUCCESS)
			return RSPF0_RTN_ERROR_SPU_API;
	}

	/* clear internal buffer in SPU2 */
	/* Sound */
	result = SPUSound_Clear(&path->sound, RSPF0_SPUAPI_WAIT_RESPONSE);
	if (result != RSPF0_RTN_SUCCESS)
		return RSPF0_RTN_ERROR_SPU_API;
	/* Buffer(In) */
	result = SPUSystem_ResetBuffer(&path->buffer[RSPF0_SOURCE],
					RSPF0_SPUAPI_WAIT_RESPONSE);
	if (result != RSPF0_RTN_SUCCESS)
		return RSPF0_RTN_ERROR_SPU_API;
	/* Buffer(Out) */
	result = SPUSystem_ResetBuffer(&path->buffer[RSPF0_DEST],
					RSPF0_SPUAPI_WAIT_RESPONSE);
	if (result != RSPF0_RTN_SUCCESS)
		return RSPF0_RTN_ERROR_SPU_API;

	/* ReStart */
	result = SPUSound_Enable(&path->sound, RSPF0_SPUAPI_WAIT_RESPONSE);
	if (result != RSPF0_RTN_SUCCESS)
		return RSPF0_RTN_ERROR_SPU_API;
	if (path->source_port == RSPF0_PORT_CPU) {
		result = SPUDataPort_Enable(&path->data_port[RSPF0_SOURCE],
						RSPF0_SPUAPI_WAIT_RESPONSE);
		if (result != RSPF0_RTN_SUCCESS)
			return RSPF0_RTN_ERROR_SPU_API;
	}
	if (path->dest_port == RSPF0_PORT_CPU) {
		result = SPUDataPort_Enable(&path->data_port[RSPF0_DEST],
						RSPF0_SPUAPI_WAIT_RESPONSE);
		if (result != RSPF0_RTN_SUCCESS)
			return RSPF0_RTN_ERROR_SPU_API;
	}
	return RSPF0_RTN_SUCCESS;
}


/*-------------------------------------------------------
Name    : rspf0_open_data_port()
Outline : creating and the initialization of the work
		area that the data port uses are done.
Input   : *path       : ptr of RSPF0_Path
			(keep setting param),
	: *param      : ptr of RSPF0_OpenParam
			(set param)
	: port_type   : I/O type(SOURCE or DET)
Output  : return code
-------------------------------------------------------*/
static long rspf0_open_data_port(RSPF0_Path *path, RSPF0_OpenParam *param,
						unsigned long port_type) {
	sSPUDataPortParam dp_param;
	long result;
	long result_del;

	/* The argument is guaranteed with RSPF0_Open() */

	/* CREATE Data Port */
	result = SPUDataPort_Create(path->dsp_id, &path->data_port[port_type]);
	if (result != RSPF0_RTN_SUCCESS)
		return RSPF0_RTN_ERROR_RESOURCE;
	/* SET PARAM Data Port */
	if (port_type == RSPF0_SOURCE)
		result = SPUDataPort_SetCallBack(&path->data_port[port_type],
					RSPF0_SourceTransferEndCallBack);
	else
		result = SPUDataPort_SetCallBack(&path->data_port[port_type],
					RSPF0_DestTransferEndCallBack);
	if (result != RSPF0_RTN_SUCCESS) {
		result = RSPF0_RTN_ERROR_RESOURCE;
		goto rspf0_open_data_port_error0;
	}

	dp_param.data_unit         = SPUDATAPORT_DATAUNIT_HALFWORD;
	dp_param.dma_transfer_mode = SPUDATAPORT_TRANSFER_MODE_BURST_ALIGNMENT3;
	dp_param.dma_swap_mode     = param->swap_mode;
	dp_param.end_notify_condition = SPUDATAPORT_END_CONDITION_DMA_END;
	if (port_type == RSPF0_SOURCE) {
		dp_param.direction = SPUDATAPORT_DIRECTION_IN;
		dp_param.dma_ch    = SPUDATAPORT_DMA_CH0;
	} else{
		dp_param.direction = SPUDATAPORT_DIRECTION_OUT;
		dp_param.dma_ch    = SPUDATAPORT_DMA_CH1;
	}
	result = SPUDataPort_SetParam(&path->data_port[port_type], &dp_param,
			&path->buffer[port_type], RSPF0_SPUAPI_WAIT_RESPONSE);
	if (result != RSPF0_RTN_SUCCESS)
		goto rspf0_open_data_port_error;
	/* ENABLE Data Port */
	result = SPUDataPort_Enable(&path->data_port[port_type],
						RSPF0_SPUAPI_WAIT_RESPONSE);
	if (result != RSPF0_RTN_SUCCESS)
		goto rspf0_open_data_port_error;

	return RSPF0_RTN_SUCCESS;

/* Error processing */
rspf0_open_data_port_error:
	result = RSPF0_RTN_ERROR_SPU_API;
rspf0_open_data_port_error0:
	result_del = SPUDataPort_Delete(&path->data_port[port_type]);
	if (result_del != RSPF0_RTN_SUCCESS)
		return RSPF0_RTN_ERROR_FATAL;		/* mistake to delete */
	return result;
}


/*-------------------------------------------------------
Name    : rspf0_close_data_port()
Outline : The work area that the data port used
		is released.
Input   : *path       : ptr of RSPF0_Path
			(keep setting param),
	: port_type   : I/O type(SOURCE or DET)
Output  : return code
-------------------------------------------------------*/
static long rspf0_close_data_port(RSPF0_Path *path, unsigned long port_type)
{
	long result;

	/* Disable */
	result = SPUDataPort_Disable(&path->data_port[port_type],
				0,	/* not read the rest of transfer */
				RSPF0_SPUAPI_WAIT_RESPONSE);
	if (result != RSPF0_RTN_SUCCESS)
		return RSPF0_RTN_ERROR_FATAL;

	/* Delete */
	/*if (path->data_port_status[port_type] == RSPF0_PORT_STATE_DISABLE)*/
		result = SPUDataPort_Delete(&path->data_port[port_type]);
	if (result != RSPF0_RTN_SUCCESS)
		return RSPF0_RTN_ERROR_FATAL;
	return RSPF0_RTN_SUCCESS;
}


/*-------------------------------------------------------
Name    : rspf0_open_fmsi_port()
Outline : creating and the initialization of the work
		area that the fmsi port uses are done.
Input   : *path       : ptr of RSPF0_Path
			(keep setting param),
	: *param      : ptr of RSPF0_OpenParam
			(set param)
	: port_type   : I/O type(SOURCE or DET)
Output  : return code
-------------------------------------------------------*/
static long rspf0_open_fmsi_port(RSPF0_Path *path, RSPF0_OpenParam *param,
						unsigned long port_type) {
	sSPUFmsiPortParam fp_param;
	unsigned long fmsi_port_id;
	long result;
	long result_del;

	const unsigned long fmsi_port_sel[2][2] = {
		{SPUFSIPORT_A_IN, SPUFSIPORT_B_IN},
		{SPUFSIPORT_A_OUT, SPUFSIPORT_B_OUT},
	};

	/* The argument is guaranteed with RSPF0_Open() */
	if (port_type == RSPF0_SOURCE)
		fmsi_port_id = fmsi_port_sel[port_type]
			[param->source_port-(unsigned long)RSPF0_PORT_FMSI_A];
	else
		fmsi_port_id = fmsi_port_sel[port_type]
			[param->dest_port-(unsigned long)RSPF0_PORT_FMSI_A];
	result = SPUFmsiPort_Create(path->dsp_id,
		/*0,*/	/* fmsi module id = 0x00(default)*/
		rspf0_fmsiip,	/* FMSI_IP_ONLY:Allocated data for FMSI IP */
		fmsi_port_id,
		&path->fmsi_port[port_type]
		);
	if (result != RSPF0_RTN_SUCCESS)
		return RSPF0_RTN_ERROR_RESOURCE;
	result = SPUFmsiPort_SetCallBack(&path->fmsi_port[port_type],
		RSPF0_FmsiPortErrorCallBack);
	if (result != RSPF0_RTN_SUCCESS) {
		result = RSPF0_RTN_ERROR_RESOURCE;
		goto rspf0_open_fmsi_port_error0;
	}
	path->fmsi_port_status[port_type] = RSPF0_PORT_STATE_DISABLE;

	/* SET PARAM FmsiPort */
	fp_param.ch = param->channel;
	fp_param.threshold_mode = 0;
	fp_param.output_start_threshold = 0;	/* Default : Half */
	result = SPUFmsiPort_SetParam(&path->fmsi_port[port_type],
	&path->buffer[port_type], &fp_param, RSPF0_SPUAPI_WAIT_RESPONSE);
	if (result != RSPF0_RTN_SUCCESS)
		goto rspf0_open_fmsi_port_error;

	/* RESET FMSI Port */
	result = SPUFmsiPort_ResetError(&path->fmsi_port[port_type],
		RSPF0_SPUAPI_WAIT_RESPONSE);
	if (result != RSPF0_RTN_SUCCESS)
		goto rspf0_open_fmsi_port_error;

	return RSPF0_RTN_SUCCESS;

/* Error processing */
rspf0_open_fmsi_port_error:
	result = RSPF0_RTN_ERROR_SPU_API;
rspf0_open_fmsi_port_error0:
	result_del = SPUFmsiPort_Delete(&path->fmsi_port[port_type]);
	if (result_del != RSPF0_RTN_SUCCESS)
		return RSPF0_RTN_ERROR_FATAL;
	return result;
}


/*-------------------------------------------------------
Name    : rspf0_close_fmsi_port()
Outline : The work area that the fmsi port used
		is released.
Input   : *path       : ptr of RSPF0_Path
			(keep setting param),
	: port_type   : I/O type(SOURCE or DET)
Output  : return code
-------------------------------------------------------*/
static long rspf0_close_fmsi_port(RSPF0_Path *path, unsigned long port_type)
{
	long result;

	/* The argument is guaranteed with RSPF0_Open() or RSPF0_Close() */
	if (path->fmsi_port_status[port_type] == RSPF0_PORT_STATE_INIT)
		return RSPF0_RTN_ERROR_FATAL;
	/* Disable */
	if (path->fmsi_port_status[port_type] == RSPF0_PORT_STATE_ENABLE) {
		result = SPUFmsiPort_Disable(&path->fmsi_port[port_type],
						RSPF0_SPUAPI_WAIT_RESPONSE);
		if (result != RSPF0_RTN_SUCCESS)
			return RSPF0_RTN_ERROR_FATAL;
		path->fmsi_port_status[port_type] = RSPF0_PORT_STATE_DISABLE;
	}
	/* Delete */
	/*if (path->fmsi_port_status[port_type] == RSPF0_PORT_STATE_DISABLE)*/
		result = SPUFmsiPort_Delete(&path->fmsi_port[port_type]);
	if (result != RSPF0_RTN_SUCCESS)
		return RSPF0_RTN_ERROR_FATAL;
	return RSPF0_RTN_SUCCESS;
}


/*-------------------------------------------------------
Name    : rspf0_open_sound()
Outline : creating and the initialization of the work
		area that the sound uses are done.
Input   : *path       : ptr of RSPF0_Path
			(keep setting param),
	: *param      : ptr of RSPF0_OpenParam
			(set param)
Output  : return code
-------------------------------------------------------*/
static long rspf0_open_sound(RSPF0_Path *path, RSPF0_OpenParam *param)
{
	sSoundParam sound_param;
	unsigned long fir_mem;
	long result;
	long result_del;

	/* The argument is guaranteed with RSPF0_Open() */
	if (param->sound.fir_mode == RSPF0_FIR_OFF)
		fir_mem = RSPF0_FIR_NONE;		/* Not Use SRC */
	else
		fir_mem = RSPF0_FIR_EXIST;		/* Use SRC */

	result = SPUSound_Create(path->dsp_id, &path->sound, fir_mem);

	if (result != RSPF0_RTN_SUCCESS)
		return RSPF0_RTN_ERROR_RESOURCE;
	if (result != RSPF0_RTN_SUCCESS) {
		result = RSPF0_RTN_ERROR_RESOURCE;
		goto rspf0_open_sound_error0;
	}

	/* SET PARAM Sound */
	sound_param.fir_mode = param->sound.fir_mode;

	result = SPUSound_SetParam(&path->sound, &path->buffer[RSPF0_SOURCE],
		&path->buffer[RSPF0_DEST], param->channel-(unsigned long)1,
		&sound_param, 64, RSPF0_SPUAPI_WAIT_RESPONSE);
	if (result != RSPF0_RTN_SUCCESS)
		goto rspf0_open_sound_error;
	result = SPUSound_Enable(&path->sound, RSPF0_SPUAPI_WAIT_RESPONSE);
	if (result != RSPF0_RTN_SUCCESS)
		goto rspf0_open_sound_error;

	return RSPF0_RTN_SUCCESS;

/* Error processing */
rspf0_open_sound_error:
	if (result == RSPF0_RTN_ERROR_PARAM)
		result = RSPF0_RTN_ERROR_PARAM;
	else
		result = RSPF0_RTN_ERROR_SPU_API;
rspf0_open_sound_error0:
	result_del = SPUSound_Delete(&path->sound);
	if (result_del != RSPF0_RTN_SUCCESS)
		return RSPF0_RTN_ERROR_FATAL;
	return result;
}


/*-------------------------------------------------------
Name    : rspf0_close_sound()
Outline : The work area that the sound used
		is released.
Input   : *path       : ptr of RSPF0_Path
			(keep setting param)
Output  : return code
-------------------------------------------------------*/
static long rspf0_close_sound(RSPF0_Path *path)
{
	long result;

	result = SPUSound_Disable(&path->sound, RSPF0_SPUAPI_WAIT_RESPONSE);
	if (result != RSPF0_RTN_SUCCESS)
		return RSPF0_RTN_ERROR_FATAL;

	/*if (path->sound_status == RSPF0_PORT_STATE_DISABLE)*/
		result = SPUSound_Delete(&path->sound);
	if (result != RSPF0_RTN_SUCCESS)
		return RSPF0_RTN_ERROR_FATAL;
	return RSPF0_RTN_SUCCESS;
}


/*-------------------------------------------------------
Name    : search_path_from_dataport()
Outline : The pointer of RSPF0_Path from pointer of
		the data port and the type of the data
		port are retrieved.
Input   : *port         : ptr of used data port,
	: port_type     : I/O type(SOURCE or DET)
Output  : index of rspf0_path. (0 is error)
-------------------------------------------------------*/
static RSPF0_Path *search_path_from_dataport(sSPUDataPort *port,
						unsigned long port_type) {
	unsigned long i;
	unsigned long dsp_id;

	if (port) {
		dsp_id = port->dsp_id;
		for (i = 0; i < RSPF0_PATH_NUM; i++) {
			if (rspf0_path[dsp_id][i]) {
				if (&(rspf0_path[dsp_id][i]->
						data_port[port_type]) == port)
					return rspf0_path[dsp_id][i];
			}
		}
	}
	return 0;				/* cannot see rspf0_path[] */
}


/*-------------------------------------------------------
Name    : RSPF0_SourceTransferEndCallBack()
Outline : The callback function of DataPort on CPU
		side is input.
Input   : *port         : ptr of used data port,
	: size          : finished size
	: flush         : flushed flag(fix 0x00)
Output  : none
-------------------------------------------------------*/
static void RSPF0_SourceTransferEndCallBack(sSPUDataPort *port,
		unsigned long size, unsigned long flush){	/* pgr0364 */
	RSPF0_Path *path;

	path = search_path_from_dataport(port, RSPF0_SOURCE);

	if (path && path->source_end_callback)
		path->source_end_callback(path, size);
	return;
}


/*-------------------------------------------------------
Name    : RSPF0_DestTransferEndCallBack()
Outline : The callback function of DataPort on CPU
		side is output.
Input   : *port         : ptr of used data port,
	: size          : finished size
	: flush         : flushed flag(non flushed=0x00,
					flushed=0x01)
Output  : none
-------------------------------------------------------*/
static void RSPF0_DestTransferEndCallBack(sSPUDataPort *port,
				unsigned long size, unsigned long flush){
	RSPF0_Path *path;

	path = search_path_from_dataport(port, RSPF0_DEST);

	if (path && path->dest_end_callback)
		path->dest_end_callback(path, size, flush);
	return;
}


/*-------------------------------------------------------
Name    : search_path_from_fmsiport()
Outline : The pointer of RSPF0_Path from pointer of
		the fmsi port and the type of the fmsi
		port are retrieved.
Input   : *port         : ptr of used data port,
Output  : index of rspf0_path. (0 is error)
-------------------------------------------------------*/
static RSPF0_Path *search_path_from_fmsiport(sSPUFmsiPort *port)
{
	unsigned long i;
	unsigned long dsp_id;

	if (port) {
		dsp_id = port->dsp_id;
		for (i = 0; i < RSPF0_PATH_NUM; i++) {
			if (rspf0_path[dsp_id][i]) {
				if (&(rspf0_path[dsp_id][i]->
					fmsi_port[RSPF0_SOURCE]) == port ||
					&(rspf0_path[dsp_id][i]->
					fmsi_port[RSPF0_DEST]) == port)
					return rspf0_path[dsp_id][i];
			}
		}
	}
	return 0;				/* cannot see rspf0_path[] */
}


/*-------------------------------------------------------
Name    : RSPF0_FmsiPortErrorCallBack()
Outline : The callback function of FMSIPort
Input   : *port         : ptr of used data port,
	: info          : information of error,
	: flush         : flushed flag(non flushed=0x00,
					flushed=0x01)
Output  : none
-------------------------------------------------------*/
static void RSPF0_FmsiPortErrorCallBack(sSPUFmsiPort *port,
				unsigned long info, unsigned long flush) {
	RSPF0_Path *path;
	unsigned long cb_info;

	path = search_path_from_fmsiport(port);

	if (path && path->fmsi_error_callback) {
		cb_info = 0;
		if (info & 0x000001)
			cb_info |= RSPF0_FMSI_INFO_UNDERFLOW;
		if (info & 0x000002)
			cb_info |= RSPF0_FMSI_INFO_OVERFLOW;
		if (flush)
			cb_info |= RSPF0_FMSI_INFO_FLUSHED;
		path->fmsi_error_callback(path, cb_info);
	}
	return;
}

/*-------------------------------------------------------
Name    : get_path_index()
Outline : Refer to index of the structure on CPU side
		from RSPF0_Path ptr.
Input   : *path           : ptr of RSPF0_Path
Output  : index of rspf0_path. (-1 is error)
-------------------------------------------------------*/
static long get_path_index(RSPF0_Path *path)
{
	long i;

	if (path) {
		for (i = 0; i < RSPF0_PATH_NUM; i++) {
			if (rspf0_path[path->dsp_id][i] == path)
				return i;
		}
	}
	return -1;					/* error index */
}

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("SPU OPEN SOUND I/F Ver 1.0");
MODULE_AUTHOR("Renesas Electronics");
